SystemOrganization addCategory: #'TrueType-Fonts'!
SystemOrganization addCategory: #'TrueType-Support'!

Object subclass: #TTContourConstruction
	instanceVariableNames: 'points'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TrueType-Support'!

!TTContourConstruction commentStamp: '<historical>' prior: 0!
This class represents a temporary contour structure during the construction of a TTGlyph from a TrueType file.

Instance variables:
	points	<Array of: TTPoint>	The points defining this contour!

!TTContourConstruction class methodsFor: 'instance creation' stamp: ''!
on: points

	^self new points: points! !

!TTContourConstruction methodsFor: 'converting' stamp: 'PeterHugossonMiller 9/3/2009 11:35'!
asCompressedPoints
	"Return the receiver compressed into a PointArray.
	All lines will be converted into bezier segments with
	the control point set to the start point"
	| out minPt maxPt fullRange |
	minPt := -16r7FFF asPoint.
	maxPt := 16r8000 asPoint.
	"Check if we need full 32bit range"
	fullRange := points anySatisfy: [:any| any asPoint < minPt or:[any asPoint > maxPt]].
	out := fullRange
		ifTrue: [(PointArray new: points size) writeStream]
		ifFalse:[(ShortPointArray new: points size) writeStream].
	self segmentsDo:[:segment|
		out nextPut: segment start.
		segment isBezier2Segment 
			ifTrue:[out nextPut: segment via]
			ifFalse:[out nextPut: segment start].
		out nextPut: segment end.
	].
	^out contents! !

!TTContourConstruction methodsFor: 'accessing' stamp: ''!
points
	^points! !

!TTContourConstruction methodsFor: 'accessing' stamp: 'ar 11/1/1998 22:34'!
points: anArray
	points := anArray asArray.! !

!TTContourConstruction methodsFor: 'printing' stamp: ''!
printOn: aStream

	aStream
		nextPutAll: self class name;
		nextPut:$(;
		print: points size;
		"space;
		print: self type;"
		nextPut:$)! !

!TTContourConstruction methodsFor: 'accessing' stamp: ''!
segments

	| segments |
	segments := OrderedCollection new.
	self segmentsDo:[:seg| segments add: seg].
	^segments! !

!TTContourConstruction methodsFor: 'enumerating' stamp: 'hmm 10/28/2001 21:55'!
segmentsDo: aBlock
	"Evaluate aBlock with the segments of the receiver. This may either be straight line
	segments or quadratic bezier curves. The decision is made upon the type flags
	in TTPoint as follows:
	a) 	Two subsequent #OnCurve points define a straight segment
	b) 	An #OnCurve point followed by an #OffCurve point followed 
		by an #OnCurve point defines a quadratic bezier segment
	c)	Two subsequent #OffCurve points have an implicitely defined 
		#OnCurve point at half the distance between them"
	| last next mid index i |
	last := points first.
	"Handle case where first point is off-curve"
	(last type == #OnCurve) ifFalse: [
		i := points findFirst: [:pt | pt type == #OnCurve].
		i = 0
			ifTrue: [mid := TTPoint new
							type: #OnCurve;
							x: points first x + points last x // 2;
							y: points first y + points last y // 2.
					points := (Array with: mid), points]
			ifFalse: [points := (points copyFrom: i to: points size), (points copyFrom: 1 to: i)].
		last := points first].
	index := 2.
	[index <= points size] whileTrue:[
		mid := points at: index.
		mid type == #OnCurve ifTrue:[
			"Straight segment"
			aBlock value: (LineSegment from: last asPoint to: mid asPoint).
			last := mid.
		] ifFalse:["Quadratic bezier"
			"Read ahead if the next point is on curve"
			next := (index < points size) ifTrue:[points at: (index+1)] ifFalse:[points first].
			next type == #OnCurve ifTrue:[
				"We'll continue after the end point"
				index := index + 1.
			] ifFalse:[ "Calculate center"
				next := (next asPoint + mid asPoint) // 2].
			aBlock value:(Bezier2Segment from: last asPoint via: mid asPoint to: next asPoint).
			last := next].
		index := index + 1].
	(index = (points size + 1)) ifTrue:[
		aBlock value:(LineSegment from: points last asPoint to: points first asPoint)]! !

Object subclass: #TTFontTableDirEntry
	instanceVariableNames: 'tag fontData offset length checkSum'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TrueType-Support'!

!TTFontTableDirEntry commentStamp: '<historical>' prior: 0!
This class represents an entry in a truetype font table directory. Used by TTFontReader only.!

!TTFontTableDirEntry class methodsFor: 'instance creation' stamp: ''!
on: fontData at: index

	^self new on: fontData at: index! !

!TTFontTableDirEntry methodsFor: 'accessing' stamp: ''!
nextByte

	| value |
	value := fontData byteAt: offset.
	offset := offset + 1.
	^value! !

!TTFontTableDirEntry methodsFor: 'accessing' stamp: ''!
nextBytes: numBytes into: array startingAt: byteOffset

	1 to: numBytes do:[:i|
		array at: i put: (fontData byteAt: byteOffset + i - 1)].! !

!TTFontTableDirEntry methodsFor: 'accessing' stamp: ''!
nextLong

	| value |
	value := fontData longAt: offset bigEndian: true.
	offset := offset + 4.
	^value! !

!TTFontTableDirEntry methodsFor: 'accessing' stamp: ''!
nextShort

	| value |
	value := fontData shortAt: offset bigEndian: true.
	offset := offset + 2.
	^value! !

!TTFontTableDirEntry methodsFor: 'accessing' stamp: ''!
nextULong

	| value |
	value := fontData unsignedLongAt: offset bigEndian: true.
	offset := offset + 4.
	^value! !

!TTFontTableDirEntry methodsFor: 'accessing' stamp: ''!
nextUShort

	| value |
	value := fontData unsignedShortAt: offset bigEndian: true.
	offset := offset + 2.
	^value! !

!TTFontTableDirEntry methodsFor: 'accessing' stamp: ''!
offset
	^offset! !

!TTFontTableDirEntry methodsFor: 'accessing' stamp: ''!
offset: newOffset
	offset := newOffset! !

!TTFontTableDirEntry methodsFor: 'initialization' stamp: ''!
on: fd at: index

	fontData := fd.
	tag := fontData longAt: index bigEndian: true.
	checkSum := fontData longAt: index+4 bigEndian: true.
	offset := (fontData longAt: index+8 bigEndian: true) + 1.
	length := fontData longAt: index+12 bigEndian: true.! !

!TTFontTableDirEntry methodsFor: 'accessing' stamp: ''!
skip: n
	"Skip n bytes"
	offset := offset + n.! !

!TTFontTableDirEntry methodsFor: 'accessing' stamp: 'ar 11/1/1998 23:37'!
stringAt: stringOffset length: byteLength multiByte: aBoolean

	| string index stringLength |
	aBoolean ifFalse:[
		stringLength := byteLength.
		string := String new: stringLength.
		index := stringOffset.
		1 to: stringLength do:[:i|
			string at: i put: (Character value: (fontData byteAt: index + i - 1))].
		^string
	] ifTrue:[
		stringLength := byteLength // 2.
		string := String new: stringLength.
		index := stringOffset.
		1 to: stringLength do:[:i|
			string at: i put: (Character value: (fontData byteAt: index + 1)).
			index := index + 2].
		^string]! !

Object subclass: #TTGlyph
	instanceVariableNames: 'bounds contours advanceWidth leftSideBearing rightSideBearing glyphIndex'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TrueType-Fonts'!

!TTGlyph commentStamp: '<historical>' prior: 0!
This class represents a glyph of a TrueType font.

Instance variables:
	bounds			<Rectangle>	The receiver's bounds
	contours		<Array of: PointArray> The compressed contours in the receiver
	advanceWidth	<Integer>	advance width of the glyph
	leftSideBearing	<Integer>	left side bearing
	rightSideBearing <Integer>	right side bearing
	glyphIndex 		<Integer>	the original index of the glyph (used for kerning)!

TTGlyph subclass: #TTCompositeGlyph
	instanceVariableNames: 'glyphs'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TrueType-Fonts'!

!TTCompositeGlyph commentStamp: '<historical>' prior: 0!
This class represents a composite TrueType glyph, e.g.one which contains many simple TTGlyphs.!

!TTCompositeGlyph methodsFor: 'accessing' stamp: 'ar 11/2/1998 01:21'!
addGlyph: aGlyph transformation: aMatrix
	glyphs := glyphs copyWith: (aMatrix -> aGlyph)! !

!TTCompositeGlyph methodsFor: 'private' stamp: 'PeterHugossonMiller 9/3/2009 11:33'!
computeContours
	| out |
	out := (Array new: glyphs size * 4) writeStream.
	self glyphsAndTransformationsDo:[:glyph :transform|
		glyph contours do:[:ptArray|
			out nextPut: (transform localPointsToGlobal: ptArray).
		].
	].
	^out contents! !

!TTCompositeGlyph methodsFor: 'accessing' stamp: 'ar 11/2/1998 01:43'!
contours
	^contours ifNil:[contours := self computeContours]! !

!TTCompositeGlyph methodsFor: 'private' stamp: 'ar 11/14/1998 20:27'!
flipAroundY
	bounds := (bounds origin x @ bounds corner y negated) corner:
				(bounds corner x @ bounds origin y negated).
	contours := nil.! !

!TTCompositeGlyph methodsFor: 'accessing' stamp: 'ar 11/2/1998 01:20'!
glyphs

	^glyphs collect:[:assoc| assoc value].! !

!TTCompositeGlyph methodsFor: 'accessing' stamp: 'ar 11/2/1998 01:28'!
glyphsAndTransformationsDo: aBlock
	glyphs do:[:assoc|
		aBlock value: assoc value value: assoc key.
	].! !

!TTCompositeGlyph methodsFor: 'initialize' stamp: 'ar 11/2/1998 01:20'!
initialize
	glyphs := #().! !

!TTCompositeGlyph methodsFor: 'testing' stamp: ''!
isComposite
	^true! !

!TTGlyph methodsFor: 'accessing' stamp: ''!
advanceWidth
	^advanceWidth! !

!TTGlyph methodsFor: 'accessing' stamp: ''!
advanceWidth: aNumber
	advanceWidth := aNumber.! !

!TTGlyph methodsFor: 'converting' stamp: 'yo 6/23/2003 18:29'!
asFormWithScale: scale ascender: ascender descender: descender
	^ self
		asFormWithScale: scale
		ascender: ascender
		descender: descender
		fgColor: Color black
		bgColor: Color white
		depth: 8
		replaceColor: true.
! !

!TTGlyph methodsFor: 'converting' stamp: 'yo 6/23/2003 18:27'!
asFormWithScale: scale ascender: ascender descender: descender fgColor: fgColor bgColor: bgColor depth: depth

	^ self
		asFormWithScale: scale
		ascender: ascender
		descender: descender
		fgColor: fgColor
		bgColor: bgColor
		depth: depth
		replaceColor: false.
! !

!TTGlyph methodsFor: 'converting' stamp: 'yo 5/7/2004 10:37'!
asFormWithScale: scale ascender: ascender descender: descender fgColor: fgColor bgColor: bgColor depth: depth replaceColor: replaceColorFlag

	^ self
		asFormWithScale: scale
		ascender: ascender
		descender: descender
		fgColor: fgColor
		bgColor: bgColor
		depth: depth
		replaceColor: replaceColorFlag
		lineGlyph: nil
		lingGlyphWidth: 0
		emphasis: 0.! !

!TTGlyph methodsFor: 'converting' stamp: 'yo 5/7/2004 11:22'!
asFormWithScale: scale ascender: ascender descender: descender fgColor: fgColor bgColor: bgColor depth: depth replaceColor: replaceColorFlag lineGlyph: lineGlyph lingGlyphWidth: lWidth emphasis: code

	| form canvas newScale |
	form := Form extent: (advanceWidth @ (ascender - descender) * scale) rounded depth: depth.
	form fillColor: bgColor.
	canvas := BalloonCanvas on: form.
	canvas aaLevel: 4.
	canvas transformBy: (MatrixTransform2x3 withScale: scale asPoint * (1 @ -1)).
	canvas transformBy: (MatrixTransform2x3 withOffset: 0 @ ascender negated).
	canvas
		drawGeneralBezierShape: self contours
		color: fgColor 
		borderWidth: 0 
		borderColor: fgColor.
	((code bitAnd: 4) ~= 0 or: [(code bitAnd: 16) ~= 0]) ifTrue: [
		newScale := (form width + 1) asFloat / lineGlyph calculateWidth asFloat.
		canvas transformBy: (MatrixTransform2x3 withScale: (newScale / scale)@1.0).

		(code bitAnd: 4) ~= 0 ifTrue: [
			canvas
				drawGeneralBezierShape: lineGlyph contours
				color: fgColor 
				borderWidth: 0 
				borderColor: fgColor.
		].

		(code bitAnd: 16) ~= 0 ifTrue: [
			canvas transformBy: (MatrixTransform2x3 withOffset: 0@(ascender // 2)).
			canvas
				drawGeneralBezierShape: lineGlyph contours
				color: fgColor 
				borderWidth: 0 
				borderColor: fgColor.
		].
	].

	replaceColorFlag ifTrue: [
		form replaceColor: bgColor withColor: Color transparent.
	].
	^ form! !

!TTGlyph methodsFor: 'accessing' stamp: 'ar 11/1/1998 22:25'!
bounds
	^bounds! !

!TTGlyph methodsFor: 'accessing' stamp: 'ar 11/1/1998 22:25'!
bounds: aRectangle
	bounds := aRectangle! !

!TTGlyph methodsFor: 'private-initialization' stamp: 'nice 1/5/2010 15:59'!
buildContours
	"Build the contours in the receiver glyph.
	The contour is constructed by converting the points
	form each contour into an absolute value and then
	compressing the contours into PointArrays."
	| tx ty |
	tx := ty := 0.
	contours := contours collect:[:contour| | points |
		points := contour points.
		points do:[:pt|
			pt x: (tx := tx + pt x).
			pt y: (ty := ty + pt y)].
		contour asCompressedPoints].! !

!TTGlyph methodsFor: 'private' stamp: 'yo 5/7/2004 10:38'!
calculateWidth

	| min max |
	min := SmallInteger maxVal.
	max := SmallInteger minVal.
	self contours do: [:a | a do: [:p |
		p x > max ifTrue: [max := p x].
		p x < min ifTrue: [min := p x].
	]].
	^ max - min.
! !

!TTGlyph methodsFor: 'accessing' stamp: ''!
contours
	^contours! !

!TTGlyph methodsFor: 'accessing' stamp: ''!
contours: aCollection
	contours := aCollection asArray.! !

!TTGlyph methodsFor: 'private' stamp: 'ar 5/25/2000 18:01'!
display
	| canvas |
	canvas := Display getCanvas.
	self contours do:[:ptArray|
		1 to: ptArray size by: 3 do:[:i|
			canvas line: (ptArray at: i) // 10
					to: (ptArray at: i+2) // 10
					width: 1 color: Color black.
		].
	].! !

!TTGlyph methodsFor: 'private' stamp: 'ar 11/14/1998 20:22'!
flipAroundY
	bounds := (bounds origin x @ bounds corner y negated) corner:
				(bounds corner x @ bounds origin y negated).
	contours := self contours collect:[:contour| contour collect:[:pt| pt x @ pt y negated]].! !

!TTGlyph methodsFor: 'accessing' stamp: ''!
glyphIndex
	^glyphIndex! !

!TTGlyph methodsFor: 'accessing' stamp: ''!
glyphIndex: anInteger
	glyphIndex := anInteger! !

!TTGlyph methodsFor: 'accessing' stamp: 'ar 11/2/1998 01:26'!
glyphsAndTransformationsDo: aBlock
	aBlock value: self value: MatrixTransform2x3 identity! !

!TTGlyph methodsFor: 'initialization' stamp: 'ar 11/1/1998 22:25'!
initialize

	bounds := 0@0 corner: 0@0.
	contours := #().
	advanceWidth := 0.
	leftSideBearing := 0.
	rightSideBearing := 0.! !

!TTGlyph methodsFor: 'private-initialization' stamp: 'ar 11/1/1998 22:42'!
initializeContours: numContours with: endPoints
	"Initialize the contours for creation of the glyph."
	| startPt pts endPt |
	contours := Array new: numContours.
	startPt := -1.
	1 to: numContours do:[:i|
		endPt := endPoints at: i.
		pts := Array new: endPt - startPt.
		1 to: pts size do:[:j| pts at: j put: TTPoint new].
		contours at: i put: (TTContourConstruction on: pts).
		startPt := endPt].! !

!TTGlyph methodsFor: 'testing' stamp: ''!
isComposite
	^false! !

!TTGlyph methodsFor: 'accessing' stamp: ''!
leftSideBearing
	^leftSideBearing! !

!TTGlyph methodsFor: 'accessing' stamp: ''!
leftSideBearing: aNumber
	leftSideBearing := aNumber.! !

!TTGlyph methodsFor: 'printing' stamp: 'tk 9/13/1999 09:54'!
printOn: aStream

	aStream
		nextPutAll: self class name;
		nextPut:$(;
		print: (contours ifNil: [0] ifNotNil: [contours size]);
		nextPut:$).! !

!TTGlyph methodsFor: 'accessing' stamp: ''!
rightSideBearing
	^rightSideBearing! !

!TTGlyph methodsFor: 'accessing' stamp: ''!
rightSideBearing: aNumber
	rightSideBearing := aNumber.! !

!TTGlyph methodsFor: 'private-initialization' stamp: 'ar 11/1/1998 22:27'!
updateRightSideBearing
	"Update the right side bearing value"
	"@@: Is the following really correct?!!?!!"
	rightSideBearing := advanceWidth - leftSideBearing - bounds corner x + bounds origin x! !

Object subclass: #TTKernPair
	instanceVariableNames: 'left right value mask'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TrueType-Fonts'!

!TTKernPair commentStamp: '<historical>' prior: 0!
A TTKernPair represents a TrueType kerning pair.

Instance variables:
	left	<Integer>	The glyph index for the left character.
	right <Integer>	The glyph index for the right character.
	value <Integer>	The amount of kerning.
	mask <Integer>	An efficient representation for the left and the right value.!

!TTKernPair class methodsFor: 'accessing' stamp: ''!
maskFor: left with: right
	^(left bitShift: 12) + right! !

!TTKernPair methodsFor: 'accessing' stamp: ''!
left
	^left! !

!TTKernPair methodsFor: 'accessing' stamp: ''!
left: aNumber

	left := aNumber! !

!TTKernPair methodsFor: 'accessing' stamp: 'ar 11/1/1998 20:08'!
mask
	^mask ifNil:[mask := self class maskFor: left with: right]! !

!TTKernPair methodsFor: 'accessing' stamp: ''!
right
	^right! !

!TTKernPair methodsFor: 'accessing' stamp: ''!
right: aNumber

	right := aNumber! !

!TTKernPair methodsFor: 'accessing' stamp: ''!
value
	^value! !

!TTKernPair methodsFor: 'accessing' stamp: ''!
value: aNumber

	value := aNumber! !

Object subclass: #TTPoint
	instanceVariableNames: 'x y type'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TrueType-Support'!

!TTPoint commentStamp: '<historical>' prior: 0!
A representation of a TrueType point which includes a 'type' flag defining whether this point is an 'on' or an 'off' curve point.!

!TTPoint methodsFor: 'converting' stamp: ''!
asPoint
	^x@y! !

!TTPoint methodsFor: 'printing' stamp: ''!
printOn: aStream

	aStream 
		nextPutAll: self class name;
		nextPut:$(;
		print: x;
		nextPut:$@;
		print: y;
		nextPut:$|;
		print: type;
		nextPut:$)! !

!TTPoint methodsFor: 'accessing' stamp: ''!
type
	^type! !

!TTPoint methodsFor: 'accessing' stamp: ''!
type: aSymbol

	type := aSymbol! !

!TTPoint methodsFor: 'accessing' stamp: ''!
x
	^x! !

!TTPoint methodsFor: 'accessing' stamp: ''!
x: aNumber

	x := aNumber! !

!TTPoint methodsFor: 'accessing' stamp: ''!
y
	^y! !

!TTPoint methodsFor: 'accessing' stamp: ''!
y: aNumber
	y := aNumber! !

Object subclass: #TTFontDescription
	instanceVariableNames: 'glyphTable glyphs kernPairs copyright familyName fullName subfamilyName uniqueName versionName postscriptName trademark bounds unitsPerEm ascender descender lineGap'
	classVariableNames: 'Default Descriptions'
	poolDictionaries: 'TextConstants'
	category: 'TrueType-Fonts'!

!TTFontDescription commentStamp: '<historical>' prior: 0!
Holds a TrueType font in memory.  Is used by TTSampleStringMorph as its font.  

Class owns a default example.  !

TTFontDescription subclass: #TTCFontDescription
	instanceVariableNames: ''
	classVariableNames: 'TTCDefault TTCDescriptions'
	poolDictionaries: 'TextConstants'
	category: 'TrueType-Fonts'!

!TTCFontDescription class methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:20'!
addFromTTFile: fileName
"
	Execute the following only if you know what you are doing.
	self addFromTTFile: 'C:\WINDOWS\Fonts\msgothic.TTC'
"

	| tt old |
	(fileName asLowercase endsWith: 'ttf') ifTrue: [
		tt := TTCFontReader readTTFFrom: (FileStream readOnlyFileNamed: fileName).
	] ifFalse: [
		tt := TTCFontReader readFrom: (FileStream readOnlyFileNamed: fileName).
	].
		
	old := TTCDescriptions detect: [:f | f first name = tt first name] ifNone: [nil].
	old ifNotNil: [TTCDescriptions remove: old].
	TTCDescriptions add: tt.
	^ tt.
! !

!TTCFontDescription class methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:20'!
clearDefault
"
	self clearDefault
"

	TTCDefault := nil.
! !

!TTCFontDescription class methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:20'!
clearDescriptions
"
	self clearDescriptions
"

	TTCDescriptions := Set new.
	TTCDefault ifNotNil: [TTCDescriptions add: TTCDefault].
! !

!TTCFontDescription class methodsFor: 'as yet unclassified' stamp: 'yo 11/30/2002 22:44'!
default

	^ TTCDefault.
! !

!TTCFontDescription class methodsFor: 'as yet unclassified' stamp: 'yo 12/10/2002 19:29'!
descriptionNamed: descriptionName

	^ TTCDescriptions detect: [:f | f first name = descriptionName] ifNone: [TTCDefault].
! !

!TTCFontDescription class methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:20'!
descriptionNamed: descriptionName at: index

	| array |
	(array :=  self descriptionNamed: descriptionName) ifNil: [^ nil].
	^ array at: index.
! !

!TTCFontDescription class methodsFor: 'as yet unclassified' stamp: 'yo 11/30/2002 23:10'!
initialize
"
	self initialize
"

	self clearDescriptions.
! !

!TTCFontDescription class methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:20'!
removeDescriptionNamed: descriptionName

	| tt |
	TTCDescriptions ifNil: [^ self].
	[(tt := TTCDescriptions detect: [:f | ('Multi', f first name) = descriptionName] ifNone: [nil]) notNil] whileTrue:[
		 TTCDescriptions remove: tt
	].
! !

!TTCFontDescription class methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:20'!
setDefault
"
	self setDefault
"

	TTCDefault := TTCFontReader readFrom: (FileStream readOnlyFileNamed: 'C:\WINDOWS\Fonts\msgothic.ttc').
	self clearDescriptions.

! !

!TTCFontDescription methodsFor: 'accessing' stamp: 'sd 2/4/2008 21:20'!
at: aCharOrInteger

	| char |
	char := aCharOrInteger asCharacter.
	^ glyphs at: (char charCode) + 1.
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'tb 6/24/2003 17:10'!
addFromTTFile: fileName
"
	self addFromTTFile: 'C:\WINDOWS\Fonts\ARIALN.TTF'
"
	^self addFromTTStream: (FileStream readOnlyFileNamed: fileName).
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'tb 6/24/2003 17:08'!
addFromTTStream: readStream
"
	self addFromTTFile: 'C:\WINDOWS\Fonts\ARIALN.TTF'
"

	| tt old |
	tt := TTFontReader readFrom: readStream.
	old := Descriptions detect: [:f | f name = tt name and: [f subfamilyName = tt subfamilyName]] ifNone: [nil].
	old ifNotNil: [Descriptions remove: old].
	Descriptions add: tt.
	^ tt.
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 11/30/2002 22:22'!
clearDefault
"
	self clearDefault
"

	Default := nil.
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 11/30/2002 22:22'!
clearDescriptions
"
	self clearDescriptions
"

	Descriptions := Set new.
	Default ifNotNil: [Descriptions add: Default].
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'tk 12/10/2001 17:12'!
default
	^ Default! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'dgd 11/4/2003 17:54'!
descriptionFullNamed: descriptionFullName 
	^ Descriptions
		detect: [:f | f fullName = descriptionFullName]
		ifNone: [Default]! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 11/30/2002 22:22'!
descriptionNamed: descriptionName

	^ Descriptions detect: [:f | f name = descriptionName] ifNone: [Default].
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 11/30/2002 22:22'!
initialize
"
	self initialize
"

	self clearDescriptions.
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 2/21/2004 02:36'!
removeDescriptionNamed: descriptionName

	| tt |
	Descriptions ifNil: [^ self].
	[(tt :=  Descriptions detect: [:f | f name = descriptionName] ifNone: [nil]) notNil] whileTrue:[
		 Descriptions remove: tt
	].
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 12/13/2002 13:55'!
removeDescriptionNamed: descriptionName subfamilyName: subfamilyName

	| tts |
	Descriptions ifNil: [^ self].
	tts := Descriptions select: [:f | f name = descriptionName and: [f subfamilyName = subfamilyName]].
	tts do: [:f | Descriptions remove: f].
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 12/13/2002 13:20'!
setDefault
"
	self setDefault
"

	Default := TTFontReader readFrom: (FileStream readOnlyFileNamed: 'C:\WINDOWS\Fonts\comic.ttf').
! !

!TTFontDescription methodsFor: 'converting' stamp: 'sma 5/5/2000 13:46'!
asStrikeFontScale: scale
	"Generate a StrikeFont (actually a FormSetFont) for this TTF font at a given scale."

	| forms |
	forms := (0 to: 255) collect:
		[:i |
		(self at: i)
			asFormWithScale: scale
			ascender: ascender
			descender: descender].
	^ FormSetFont new
		fromFormArray: forms
		asciiStart: 0
		ascent: (ascender * scale) rounded! !

!TTFontDescription methodsFor: 'properties' stamp: 'ar 11/14/1998 23:48'!
ascender
	^ascender! !

!TTFontDescription methodsFor: 'accessing' stamp: 'ar 11/2/1998 01:08'!
at: aCharOrInteger
	^glyphTable at: aCharOrInteger asInteger+1! !

!TTFontDescription methodsFor: 'accessing' stamp: 'ar 11/2/1998 01:08'!
at: index put: value
	^self shouldNotImplement! !

!TTFontDescription methodsFor: 'migration' stamp: 'yo 8/16/2004 10:44'!
blankGlyphForSeparators

	| space |
	space := (self at: Character space charCode) copy.
	space contours: #().
	Character separators do: [:s | 
		glyphTable at: s charCode +1 put: space.
	].
! !

!TTFontDescription methodsFor: 'properties' stamp: 'ar 11/14/1998 23:48'!
bounds
	^bounds! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:48'!
copyright
	^copyright! !

!TTFontDescription methodsFor: 'copying' stamp: 'yo 6/23/2003 18:23'!
deepCopy

	"Since it shouldn't be copied for transmitting or any reason, it returns self."
	^ self.
! !

!TTFontDescription methodsFor: 'properties' stamp: 'ar 11/14/1998 23:48'!
descender
	^descender! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:48'!
familyName
	^familyName! !

!TTFontDescription methodsFor: 'private-initialization' stamp: 'ar 11/14/1998 20:20'!
flipAroundY
	bounds := (bounds origin x @ bounds corner y negated) corner:
				(bounds corner x @ bounds origin y negated).
	glyphs do:[:glyph| glyph flipAroundY]! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:48'!
fullName
	^fullName! !

!TTFontDescription methodsFor: 'properties' stamp: 'ar 11/14/1998 23:48'!
lineGap
	^lineGap! !

!TTFontDescription methodsFor: 'accessing' stamp: 'yo 11/30/2002 22:38'!
name

	^ self familyName copyWithout: Character space.
! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:48'!
postscriptName
	^postscriptName! !

!TTFontDescription methodsFor: 'printing' stamp: 'th 6/27/2003 17:08'!
printOn: aStream
	super printOn: aStream.
	aStream nextPut: $(.
	familyName printOn: aStream.
	aStream nextPut:$).! !

!TTFontDescription methodsFor: 'private-initialization' stamp: 'ar 11/2/1998 00:27'!
setAscender: asc descender: desc lineGap: lgap
	ascender := asc.
	descender := desc.
	lineGap := lgap! !

!TTFontDescription methodsFor: 'private-initialization' stamp: 'ar 11/2/1998 00:28'!
setBounds: aRect unitsPerEm: aNumber
	bounds := aRect.
	unitsPerEm := aNumber.! !

!TTFontDescription methodsFor: 'private-initialization' stamp: 'ar 11/2/1998 00:27'!
setGlyphs: glyphArray mapping: mappingTable
	glyphs := glyphArray.
	glyphTable := mappingTable.! !

!TTFontDescription methodsFor: 'private-initialization' stamp: 'ar 11/2/1998 00:48'!
setKernPairs: array
	kernPairs := array! !

!TTFontDescription methodsFor: 'private-initialization' stamp: 'ar 11/2/1998 00:46'!
setStrings: anArray
	copyright := anArray at: 1.
	familyName := anArray at: 2.
	subfamilyName := anArray at: 3.
	uniqueName := anArray at: 4.
	fullName := anArray at: 5.
	versionName := anArray at: 6.
	postscriptName := anArray at: 7.
	trademark := anArray at: 8.
! !

!TTFontDescription methodsFor: 'accessing' stamp: 'yo 11/30/2002 22:38'!
size

	^ glyphs size.
! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:48'!
subfamilyName
	^subfamilyName! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:48'!
trademark
	^trademark! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:48'!
uniqueName
	^uniqueName! !

!TTFontDescription methodsFor: 'properties' stamp: 'ar 11/14/1998 23:49'!
unitsPerEm
	^unitsPerEm! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:49'!
versionName
	^versionName! !

!TTFontDescription methodsFor: 'copying' stamp: 'yo 11/30/2002 22:38'!
veryDeepCopyWith: deepCopier
	"Return self.  I am shared.  Do not record me."
! !

Object subclass: #TTFontReader
	instanceVariableNames: 'charMap glyphs nGlyphs kernPairs fontDescription'
	classVariableNames: ''
	poolDictionaries: 'TextConstants'
	category: 'TrueType-Support'!

!TTFontReader commentStamp: '<historical>' prior: 0!
TTFontReader constructs a TTFontDescription from a TrueType font (.ttf).!

TTFontReader subclass: #TTCFontReader
	instanceVariableNames: 'fonts'
	classVariableNames: 'EncodingTag'
	poolDictionaries: ''
	category: 'TrueType-Support'!

!TTCFontReader class methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:20'!
encodingTag: aNumber
"
	TTCFontReader encodingTag: 6
"

	EncodingTag := aNumber.
! !

!TTCFontReader methodsFor: 'as yet unclassified' stamp: 'nice 1/5/2010 15:59'!
decodeCmapFmtTable: entry
	| cmapFmt length entryCount segCount segments offset cmap firstCode |
	cmapFmt := entry nextUShort.
	length := entry nextUShort.
	entry skip: 2. "skip version"

	cmapFmt = 0 ifTrue: "byte encoded table"
		[length := length - 6. 		"should be always 256"
		length <= 0 ifTrue: [^ nil].	"but sometimes, this table is empty"
		cmap := Array new: length.
		entry nextBytes: length into: cmap startingAt: entry offset.
		^ cmap].

	cmapFmt = 4 ifTrue: "segment mapping to deltavalues"
		[segCount := entry nextUShort // 2.
		entry skip: 6. "skip searchRange, entrySelector, rangeShift"
		segments := Array new: segCount.
		segments := (1 to: segCount) collect: [:e | Array new: 4].
		1 to: segCount do: [:i | (segments at: i) at: 2 put: entry nextUShort]. "endCount"
		entry skip: 2. "skip reservedPad"
		1 to: segCount do: [:i | (segments at: i) at: 1 put: entry nextUShort]. "startCount"
		1 to: segCount do: [:i | (segments at: i) at: 3 put: entry nextShort]. "idDelta"
		offset := entry offset.
		1 to: segCount do: [:i | (segments at: i) at: 4 put: entry nextUShort]. "idRangeOffset"
		cmap := Array new: 65536 withAll: 0.
		segments withIndexDo:
			[:seg :si | | code |
			seg first to: seg second do:
				[:i |
					seg last > 0 ifTrue:
						["offset to glypthIdArray - this is really C-magic!!"
						entry offset: i - seg first - 1 * 2 + seg last + si + si + offset.
						code := entry nextUShort.
						code > 0 ifTrue: [code := code + seg third]]
					ifFalse:
						["simple offset"
						code := i + seg third].
					cmap at: i + 1 put: (code \\ 16r10000)]].
		^ cmap].

	cmapFmt = 6 ifTrue: "trimmed table"
		[firstCode := entry nextUShort.
		entryCount := entry nextUShort.
		cmap := Array new: entryCount + firstCode withAll: 0.
		entryCount timesRepeat:
			[cmap at: (firstCode := firstCode + 1) put: entry nextUShort].
		^ cmap].
	^ nil! !

!TTCFontReader methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:20'!
getTableDirEntry: tagString from: fontData offset: offset
	"Find the table named tagString in fontData and return a table directory entry for it."

	| nTables pos currentTag tag |
	nTables := fontData shortAt: 5 + offset bigEndian: true.
	tag := ByteArray new: 4.
	1 to: 4 do:[:i| tag byteAt: i put: (tagString at: i) asInteger].
	tag := tag longAt: 1 bigEndian: true.
	pos := 13 + offset.
	1 to: nTables do:[:i|
		currentTag := fontData longAt: pos bigEndian: true.
		currentTag = tag ifTrue:[^TTFontTableDirEntry on: fontData at: pos].
		pos := pos+16].
	^nil! !

!TTCFontReader methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:20'!
parseTTCHeaderFrom: fontData

	| pos nTables |
	nTables := fontData longAt: 9 bigEndian: true.
	fonts := Array new: nTables.
	pos := 13.
	1 to: nTables do: [:i |
		fonts at: i put: (fontData longAt: pos bigEndian: true).
		pos := pos + 4.
	].

	^ fonts
! !

!TTCFontReader methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:20'!
processCharMap: assoc
	"Process the given character map"

	| glyph cmap encode0 encode1 char value null |
	cmap := assoc value.
	null := (glyphs at: (cmap at: Character space asUnicode + 1) + 1) copy.
	null contours: #().

	encode0 := Array new: 256 withAll: glyphs first.
	encode1 := Array new: 65536 withAll: glyphs first.

	0 to: 255 do: [:i |
		char := Character value: i.
		glyph := glyphs at: (cmap at: char asUnicode + 1) + 1.
		encode0 at: i+1 put: glyph.
	].
	Character separators do: [:c |
		encode0 at: (c asciiValue + 1) put: null.
	].
	0 to: 65536 - 1 do: [:i |
		value := cmap at: i+1.
		value = 65535 ifFalse: [ "???"
			encode1 at: i+1 put: (glyphs at: value+1).
		]
	].

	^ {encode0. encode1}.
! !

!TTCFontReader methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:20'!
processCharacterMappingTable: entry
	"Read the font's character to glyph index mapping table.
	If an appropriate mapping can be found then return an association
	with the format identifier and the contents of the table"
	| copy initialOffset nSubTables pID sID offset cmap assoc |
	initialOffset := entry offset.
	entry skip: 2. "Skip table version"
	nSubTables := entry nextUShort.
	1 to: nSubTables do:[:i|
		pID := entry nextUShort.
		sID := entry nextUShort.
		offset := entry nextULong.
		"Check if this is either a Macintosh encoded table
		or a Windows encoded table"
		(pID = 1 or:[pID = 3]) ifTrue:[
			"Go to the beginning of the table"
			copy := entry copy.
			copy offset: initialOffset + offset.
			cmap := self decodeCmapFmtTable: copy.
			"(pID = 1 and: [cmap notNil])" "Prefer Macintosh encoding over everything else"
				"ifTrue: [pID -> cmap]."
			assoc := pID -> cmap. "Keep it in case we don't find a Mac encoded table"
		].
	].
	^assoc! !

!TTCFontReader methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:20'!
readFrom: aStream

	"Read the raw font byte data"
	| fontData |
	(aStream respondsTo: #binary) ifTrue:[aStream binary].
	fontData := aStream contents asByteArray.

	fonts := self parseTTCHeaderFrom: fontData.
	^ ((Array with: fonts first) collect: [:offset |
		fontDescription := TTCFontDescription new.
		self readFrom: fontData fromOffset: offset at: EncodingTag.
	]) at: 1.
! !

!TTCFontReader methodsFor: 'as yet unclassified' stamp: 'StephaneDucasse 11/7/2011 22:45'!
readFrom: fontData fromOffset: offset at: encodingTag
	| headerEntry maxProfileEntry nameEntry indexLocEntry charMapEntry glyphEntry horzHeaderEntry horzMetricsEntry kerningEntry glyphOffset cmap numHMetrics indexToLocFormat fontDescription0 fontDescription1 array result |
	"Search the tables required to build the font"
	(headerEntry := self getTableDirEntry: 'head' from: fontData offset: offset) == nil
		ifTrue: [ ^ self error: 'This font does not have a header table' ].
	(maxProfileEntry := self getTableDirEntry: 'maxp' from: fontData offset: offset) == nil
		ifTrue: [ ^ self error: 'This font does not have a maximum profile table' ].
	(nameEntry := self getTableDirEntry: 'name' from: fontData offset: offset) == nil
		ifTrue: [ ^ self error: 'This font does not have a name table' ].
	(indexLocEntry := self getTableDirEntry: 'loca' from: fontData offset: offset) == nil
		ifTrue: [ ^ self error: 'This font does not have a relocation table' ].
	(charMapEntry := self getTableDirEntry: 'cmap' from: fontData offset: offset) == nil
		ifTrue: [ ^ self error: 'This font does not have a character map table' ].
	(glyphEntry := self getTableDirEntry: 'glyf' from: fontData offset: offset) == nil
		ifTrue: [ ^ self error: 'This font does not have a glyph table' ].
	(horzHeaderEntry := self getTableDirEntry: 'hhea' from: fontData offset: offset) == nil
		ifTrue: [ ^ self error: 'This font does not have a horizontal header table' ].
	(horzMetricsEntry := self getTableDirEntry: 'hmtx' from: fontData offset: offset) == nil
		ifTrue: [ ^ self error: 'This font does not have a horizontal metrics table' ].
	(kerningEntry := self getTableDirEntry: 'kern' from: fontData offset: offset) == nil
		ifTrue: [ 
			self crTrace: 'This font does not have a kerning table' ].	"Process the data"
	indexToLocFormat := self processFontHeaderTable: headerEntry.
	self processMaximumProfileTable: maxProfileEntry.
	self processNamingTable: nameEntry.
	glyphOffset := self processIndexToLocationTable: indexLocEntry format: indexToLocFormat.
	cmap := self processCharacterMappingTable: charMapEntry.
	(cmap == nil or: [ cmap value == nil ])
		ifTrue: [ ^ self error: 'This font has no suitable character mappings' ].
	self processGlyphDataTable: glyphEntry offsets: glyphOffset.
	numHMetrics := self processHorizontalHeaderTable: horzHeaderEntry.
	self processHorizontalMetricsTable: horzMetricsEntry length: numHMetrics.
	kerningEntry isNil
		ifTrue: [ kernPairs := #() ]
		ifFalse: [ self processKerningTable: kerningEntry ].
	array := self processCharMap: cmap.
	fontDescription0 := fontDescription shallowCopy.
	fontDescription1 := fontDescription shallowCopy.
	fontDescription0 setGlyphs: (array at: 1) mapping: nil.
	fontDescription1 setGlyphs: (array at: 2) mapping: nil.	"fontDescription setKernPairs: kernPairs."
	result := OrderedCollection new.
	(encodingTag isNil or: [ encodingTag = 0 ])
		ifTrue: [ ^ Array with: fontDescription1 ].
	result add: fontDescription0.
	encodingTag - 1 timesRepeat: [ result add: nil ].
	result add: fontDescription1.
	^ result asArray! !

!TTCFontReader methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:20'!
readTTFFrom: aStream

	"Read the raw font byte data"
	| fontData |
	(aStream respondsTo: #binary) ifTrue:[aStream binary].
	fontData := aStream contents asByteArray.
	fontDescription := TTCFontDescription new.

	^ self readFrom: fontData fromOffset: 0 at: EncodingTag.
! !

!TTFontReader class methodsFor: 'initialization' stamp: 'nk 7/16/2003 15:56'!
fileReaderServicesForFile: fullName suffix: suffix


	^(suffix = 'fnt')  | (suffix = '*') 
		ifTrue: [ self services]
		ifFalse: [#()]
! !

!TTFontReader class methodsFor: 'initialization' stamp: 'MarcusDenker 5/27/2011 14:41'!
initialize
	"self initialize"

	Smalltalk tools fileList registerFileReader: self! !

!TTFontReader class methodsFor: 'instance creation' stamp: 'BenjaminVanRyseghem 11/28/2010 19:02'!
installTTF: ttfFileName asTextStyle: textStyleName sizes: sizeArray
	"Sizes are in pixels."
	"TTFontReader
		installTTF: 'F:\fonts\amazon:=:=.TTF' 
		asTextStyle: #Amazon
		sizes: #(24 60)"

	| ttf fontArray |
	ttf := self parseFileNamed: ttfFileName.
	fontArray := sizeArray collect:
		[:each |
		(ttf asStrikeFontScale: each / ttf unitsPerEm)
			name: textStyleName;
			pointSize: each].
	TextSharedInformation at: textStyleName asSymbol put: (TextStyle fontArray: fontArray)! !

!TTFontReader class methodsFor: 'initialization' stamp: 'sd 2/1/2002 19:31'!
openTTFFile: fullName 

	(TTFontReader parseFileNamed: fullName) asMorph open! !

!TTFontReader class methodsFor: 'instance creation' stamp: 'PeterHugossonMiller 9/2/2009 16:01'!
parseFileNamed: aString
	"TTFontReader parseFileNamed:'c:\windows\arial.ttf'"
	"TTFontReader parseFileNamed:'c:\windows\times.ttf'"
	
	| contents |
	contents := (FileStream readOnlyFileNamed: aString) binary contentsOfEntireFile.
	^self readFrom: contents readStream.! !

!TTFontReader class methodsFor: 'instance creation' stamp: 'ar 11/2/1998 00:53'!
readFrom: aStream

	^self new readFrom: aStream! !

!TTFontReader class methodsFor: 'instance creation' stamp: 'yo 2/15/2004 18:40'!
readTTFFrom: aStream

	^self new readTTFFrom: aStream! !

!TTFontReader class methodsFor: 'initialization' stamp: 'tbn 8/11/2010 10:21'!
serviceOpenTrueTypeFont

	^ SimpleServiceEntry 
				provider: self 
				label: 'Open true type font'
				selector: #openTTFFile:
				description: 'Open true type font'! !

!TTFontReader class methodsFor: 'initialization' stamp: 'sd 2/1/2002 22:03'!
services

	^ Array with: self serviceOpenTrueTypeFont
! !

!TTFontReader class methodsFor: 'initialization' stamp: 'IgorStasenko 4/26/2011 16:21'!
unload

	Smalltalk tools fileList unregisterFileReader: self ! !

!TTFontReader methodsFor: 'private' stamp: 'nice 1/5/2010 15:59'!
decodeCmapFmtTable: entry
	| cmapFmt length entryCount segCount segments offset cmap firstCode |
	cmapFmt := entry nextUShort.
	length := entry nextUShort.
	entry skip: 2. "skip version"

	cmapFmt = 0 ifTrue: "byte encoded table"
		[length := length - 6. 		"should be always 256"
		length <= 0 ifTrue: [^ nil].	"but sometimes, this table is empty"
		cmap := Array new: length.
		entry nextBytes: length into: cmap startingAt: entry offset.
		^ cmap].

	cmapFmt = 4 ifTrue: "segment mapping to deltavalues"
		[segCount := entry nextUShort // 2.
		entry skip: 6. "skip searchRange, entrySelector, rangeShift"
		segments := Array new: segCount.
		segments := (1 to: segCount) collect: [:e | Array new: 4].
		1 to: segCount do: [:i | (segments at: i) at: 2 put: entry nextUShort]. "endCount"
		entry skip: 2. "skip reservedPad"
		1 to: segCount do: [:i | (segments at: i) at: 1 put: entry nextUShort]. "startCount"
		1 to: segCount do: [:i | (segments at: i) at: 3 put: entry nextShort]. "idDelta"
		offset := entry offset.
		1 to: segCount do: [:i | (segments at: i) at: 4 put: entry nextUShort]. "idRangeOffset"
		cmap := Array new: 256 withAll: 0. "could be larger, but Squeak can't handle that"
		segments withIndexDo:
			[:seg :si | | code |
			seg first to: seg second do:
				[:i |
				i < 256 ifTrue:
					[seg last > 0 ifTrue:
						["offset to glypthIdArray - this is really C-magic!!"
						entry offset: i - seg first - 1 * 2 + seg last + si + si + offset. 
						code := entry nextUShort.
						code > 0 ifTrue: [code := code + seg third]]
					ifFalse:
						["simple offset"
						code := i + seg third].
					cmap at: i + 1 put: code]]].
		^ cmap].

	cmapFmt = 6 ifTrue: "trimmed table"
		[firstCode := entry nextUShort.
		entryCount := entry nextUShort.
		cmap := Array new: entryCount + firstCode withAll: 0.
		entryCount timesRepeat:
			[cmap at: (firstCode := firstCode + 1) put: entry nextUShort].
		^ cmap].
	^ nil! !

!TTFontReader methodsFor: 'private' stamp: 'ar 11/2/1998 01:33'!
getGlyphFlagsFrom: entry size: nPts
	"Read in the flags for this glyph.  The outer loop gathers the flags that
	are actually contained in the table.  If the repeat bit is set in a flag
	then the next byte is read from the table; this is the number of times
	to repeat the last flag.  The inner loop does this, incrementing the
	outer loops index each time."
	| flags index repCount flagBits |
	flags := ByteArray new: nPts.
	index := 1.
	[index <= nPts] whileTrue:[
		flagBits := entry nextByte.
		flags at: index put: flagBits.
		(flagBits bitAnd: 8) = 8 ifTrue:[
			repCount := entry nextByte.
			repCount timesRepeat:[
				index := index + 1.
				flags at: index put: flagBits]].
		index := index + 1].
	^flags! !

!TTFontReader methodsFor: 'private' stamp: 'ar 11/2/1998 01:33'!
getTableDirEntry: tagString from: fontData
	"Find the table named tagString in fontData and return a table directory entry for it."
	| nTables pos currentTag tag |
	nTables := fontData shortAt: 5 bigEndian: true.
	tag := ByteArray new: 4.
	1 to: 4 do:[:i| tag byteAt: i put: (tagString at: i) asInteger].
	tag := tag longAt: 1 bigEndian: true.
	pos := 13.
	1 to: nTables do:[:i|
		currentTag := fontData longAt: pos bigEndian: true.
		currentTag = tag ifTrue:[^TTFontTableDirEntry on: fontData at: pos].
		pos := pos+16].
	^nil! !

!TTFontReader methodsFor: 'private' stamp: 'michael.rueger 2/5/2009 17:02'!
macToWin: index
	^(index - 1) asCharacter macRomanToUnicode charCode + 1! !

!TTFontReader methodsFor: 'processing' stamp: 'th 6/27/2003 16:58'!
processCharMap: assoc
	"Process the given character map"

	| charTable glyph cmap |
	cmap := assoc value.
	charTable := Array new: 256 withAll: glyphs first. "Initialize with default glyph"

	assoc key = 1 ifTrue: "Mac encoded table"
		[1 to: (cmap size min: charTable size) do:
			[:i |
			glyph := glyphs at: (cmap at: i) + 1.
			charTable at: (self macToWin: i) put: glyph]].

	assoc key = 3 ifTrue: "Win encoded table"
		[1 to: (cmap size min: charTable size) do:
			[:i |
			glyph := glyphs at: (cmap at: i) + 1.
			charTable at: i put: glyph]].

	^ charTable! !

!TTFontReader methodsFor: 'processing' stamp: 'yo 6/29/2004 23:33'!
processCharacterMappingTable: entry
	"Read the font's character to glyph index mapping table.
	If an appropriate mapping can be found then return an association
	with the format identifier and the contents of the table"
	| copy initialOffset nSubTables pID sID offset cmap assoc |
	initialOffset := entry offset.
	entry skip: 2. "Skip table version"
	nSubTables := entry nextUShort.
	1 to: nSubTables do:[:i|
		pID := entry nextUShort.
		sID := entry nextUShort.
		offset := entry nextULong.
		"Check if this is either a Macintosh encoded table
		or a Windows encoded table"
		(pID = 1 or:[pID = 3]) ifTrue:[
			"Go to the beginning of the table"
			copy := entry copy.
			copy offset: initialOffset + offset.
			cmap := self decodeCmapFmtTable: copy.
			(pID = 3 and: [cmap notNil]) "Prefer Windows encoding over everything else"
				ifTrue: [^ pID -> cmap].
			assoc := pID -> cmap. "Keep it in case we don't find a Mac encoded table"
		].
	].
	^assoc! !

!TTFontReader methodsFor: 'processing' stamp: 'MarcusDenker 3/27/2011 09:21'!
processCompositeGlyph: glyph contours: nContours from: entry
	"Read a composite glyph from the font data. The glyph passed into this method contains some state variables that must be copied into the resulting composite glyph."
	| flags glyphIndex hasInstr cGlyph ofsX ofsY iLen a11 a12 a21 a22 m |
	cGlyph := TTCompositeGlyph new.
	a11 := a22 := 16r4000.	"1.0 in F2Dot14"
	a21 := a12 := 0.		"0.0 in F2Dot14"
	"Copy state"
	cGlyph bounds: glyph bounds; glyphIndex: glyph glyphIndex.
	hasInstr := false.
	[ flags := entry nextUShort.
	glyphIndex := entry nextUShort + 1.
	(flags bitAnd: 1) = 1 ifTrue:[
		ofsX := entry nextShort.
		ofsY := entry nextShort.
	] ifFalse:[
		(ofsX := entry nextByte) > 127 ifTrue:[ofsX := ofsX - 256].
		(ofsY := entry nextByte) > 127 ifTrue:[ofsY := ofsY - 256]].
	((flags bitAnd: 2) = 2) ifFalse:[self error: 'Reading the values as points is not supported'].
	(flags bitAnd: 8) = 8 ifTrue:[
		a11 := a22 := entry nextShort].
	(flags bitAnd: 64) = 64 ifTrue:[
		a11 := entry nextShort.
		a22 := entry nextShort].
	(flags bitAnd: 128) = 128 ifTrue:[
		"2x2 transformation"
		a11 := entry nextShort.
		a21 := entry nextShort.
		a12 := entry nextShort.
		a22 := entry nextShort].
	m := MatrixTransform2x3 new.
	"Convert entries from F2Dot14 to float"
	m a11: (a11 asFloat / 16r4000).
	m a12: (a12 asFloat / 16r4000).
	m a21: (a21 asFloat / 16r4000).
	m a22: (a22 asFloat / 16r4000).
	m a13: ofsX.
	m a23: ofsY.
	cGlyph addGlyph: (glyphs at: glyphIndex) transformation: m.
	hasInstr := hasInstr or:[ (flags bitAnd: 256) = 256].
	"Continue as long as the MORE:=COMPONENTS bit is set"
	(flags bitAnd: 32) = 32] whileTrue.
	hasInstr ifTrue:[
		iLen := entry nextUShort.
		entry skip: iLen].
	^cGlyph! !

!TTFontReader methodsFor: 'processing' stamp: 'ar 11/2/1998 00:42'!
processFontHeaderTable: entry
"Value				Data Type    Description
unitsPerEm			USHORT      Granularity of the font's em square.
xMax				USHORT      Maximum X-coordinate for the entire font.
xMin				USHORT      Minimum X-coordinate for the entire font.
yMax				USHORT      Maximum Y-coordinate for the entire font.
yMin				USHORT      Minimum Y-coordinate for the entire font.
indexToLocFormat	SHORT       Used when processing the Index To Loc Table."
	| origin corner units indexToLocFormat |
	entry skip: 4. "Skip table version number"
	entry skip: 4. "Skip font revision number"
	entry skip: 4. "Skip check sum adjustment"
	entry skip: 4. "Skip magic number"
	entry skip: 2. "Skip flags"

	units := entry nextUShort.

	entry skip: 8. "Skip creation date"
	entry skip: 8. "Skip modification date"

	"Get min/max values of all glyphs"
	origin := entry nextShort @ entry nextShort.
	corner := entry nextShort @ entry nextShort.

	entry skip: 2. "Skip mac style"
	entry skip: 2. "Skip lowest rec PPEM"
	entry skip: 2. "Skip font direction hint"
	indexToLocFormat := entry nextShort.

	fontDescription setBounds: (origin corner: corner) unitsPerEm: units.
	^indexToLocFormat! !

!TTFontReader methodsFor: 'processing' stamp: 'SeanDeNigris 6/21/2012 09:15'!
processGlyphDataTable: entry offsets: offsetArray
	"Read the actual glyph data from the font.
	offsetArray contains the start offsets in the data for each glyph."
	| initialOffset |
	initialOffset := entry offset.
	glyphs := Array new: nGlyphs.
	1 to: nGlyphs do:[:i | 
		glyphs at: i put: (TTGlyph new glyphIndex: i-1)].
	'Reading glyph data' 
		displayProgressFrom: 1 to: nGlyphs 
		during: [:bar| | origin glyphLength corner glyphOffset nContours glyph nextOffset |
					1 to: nGlyphs do: [:glyphIndex |
							bar current: glyphIndex.
							glyph := glyphs at: glyphIndex.
							glyphOffset := offsetArray at: glyphIndex.
							nextOffset := offsetArray at: glyphIndex+1.
							glyphLength := nextOffset - glyphOffset.
					glyphLength = 0 ifFalse: [
							entry offset: initialOffset + glyphOffset.
							nContours := entry nextShort.
							origin := entry nextShort @ entry nextShort.
							corner := entry nextShort @ entry nextShort.
							glyph bounds: (origin corner: corner).
					nContours >= 0 ifTrue: [
							self processSimpleGlyph: glyph contours: nContours from: entry
				] ifFalse:[
				glyph := self processCompositeGlyph: glyph contours: nContours from: entry.
				glyphs at: glyphIndex put: glyph]]]
	].! !

!TTFontReader methodsFor: 'processing' stamp: 'ar 11/2/1998 00:40'!
processHorizontalHeaderTable: entry
"
ascender           SHORT          Typographic ascent.
descender          SHORT          Typographic descent.
lineGap            SHORT          Typographic lineGap.
numberOfHMetrics   USHORT         Number hMetric entries in the HTMX
                                               Table; may be smaller than the total
                                             number of glyphs.
"
	| asc desc lGap numHMetrics |
	entry skip: 4. "Skip table version"
	asc := entry nextShort.
	desc := entry nextShort.
	lGap := entry nextShort.
	entry skip: 2. "Skip advanceWidthMax"
	entry skip: 2. "Skip minLeftSideBearing"
	entry skip: 2. "Skip minRightSideBearing"
	entry skip: 2. "Skip xMaxExtent"
	entry skip: 2. "Skip caretSlopeRise"
	entry skip: 2. "Skip caretSlopeRun"
	entry skip: 10. "Skip 5 reserved shorts"
	entry skip: 2. "Skip metricDataFormat"

	numHMetrics := entry nextUShort.

	fontDescription setAscender: asc descender: desc lineGap: lGap.
	^numHMetrics! !

!TTFontReader methodsFor: 'processing' stamp: 'ar 11/2/1998 00:40'!
processHorizontalMetricsTable: entry length: numHMetrics
	"Extract the advance width, left side bearing, and right
	side bearing for each glyph from the Horizontal Metrics Table."
	|  index lastAW glyph |
	index := 1.
	[index <= numHMetrics] whileTrue:[
		glyph := glyphs at: index.
		glyph advanceWidth: entry nextUShort.
		glyph leftSideBearing: entry nextShort.
		glyph updateRightSideBearing.
		index := index + 1].
	index = (nGlyphs +1) ifTrue:[^true].
	lastAW := (glyphs at: index-1) advanceWidth.

	[index <= nGlyphs] whileTrue:[
		glyph := glyphs at: index.
		glyph advanceWidth: lastAW.
		glyph leftSideBearing: entry nextShort.
		glyph updateRightSideBearing.
		index := index + 1].! !

!TTFontReader methodsFor: 'processing' stamp: 'ar 11/2/1998 00:43'!
processIndexToLocationTable: entry format: indexToLocFormat
"glyphOffset    ULONG[numGlyphs]   An array that contains each glyph's
                                 offset into the Glyph Data Table.
"	| glyphOffset offset|
	glyphOffset := Array new: nGlyphs+1.
	1 to: nGlyphs+1 do:[:i|
		(indexToLocFormat = 0) ifTrue:[ "Format0: offset/2 is stored"
			offset := entry nextUShort * 2.
		] ifFalse:["Format1: store actual offset"
			offset := entry nextULong].
		glyphOffset at: i put: offset].
	^glyphOffset! !

!TTFontReader methodsFor: 'processing' stamp: 'ar 11/1/1998 23:21'!
processKerningTable: entry
	"Extract the kerning information for pairs of glyphs."
	| covLow covHigh nKernPairs kp |
	entry skip: 2. "Skip table version"
	entry skip: 2. "Skip number of sub tables -- we're using the first one only"
	entry skip: 2. "Skip current subtable number"
	entry skip: 2. "Skip length of subtable"
	covHigh := entry nextByte.
	covLow := entry nextByte.

	"Make sure the format is right (kerning table and format type 0)"
	((covLow bitAnd: 2) = 2 or:[ covHigh ~= 0]) ifTrue:[^false].
	nKernPairs := entry nextUShort.
	entry skip: 2. "Skip search range"
	entry skip: 2. "Skip entry selector"
	entry skip: 2. "Skip range shift"
	kernPairs := Array new: nKernPairs.
	1 to: nKernPairs do:[:i|
		kp := TTKernPair new.
		kp left: entry nextUShort.
		kp right: entry nextUShort.
		kp value: entry nextShort.
		kernPairs at: i put: kp].
	^true! !

!TTFontReader methodsFor: 'processing' stamp: ''!
processMaximumProfileTable: entry
"
numGlyphs         USHORT      The number of glyphs in the font.
"
	entry skip: 4. "Skip Table version number"
	nGlyphs := entry nextUShort.! !

!TTFontReader methodsFor: 'processing' stamp: 'ar 11/2/1998 00:38'!
processNamingTable: entry
"copyright         CHARPTR     The font's copyright notice.
familyName        CHARPTR     The font's family name.
subfamilyName     CHARPTR     The font's subfamily name.
uniqueName        CHARPTR     A unique identifier for this font.
fullName          CHARPTR     The font's full name (a combination of
                                          familyName and subfamilyName).
versionName       CHARPTR     The font's version string.
"	| nRecords initialOffset storageOffset pID sID lID nID length offset multiBytes string strings |
	strings := Array new: 8.
	strings atAllPut:''.
	initialOffset := entry offset.
	entry skip: 2. "Skip format selector"
	"Get the number of name records"
	nRecords := entry nextUShort.
	"Offset from the beginning of this table"
	storageOffset := entry nextUShort + initialOffset.
	1 to: nRecords do:[:i|
		pID := entry nextUShort.
		sID := entry nextUShort.
		lID := entry nextUShort.
		nID := entry nextUShort.
		length := entry nextUShort.
		offset := entry nextUShort.
		"Read only Macintosh or Microsoft strings"
		(pID = 1 or:[pID = 3 and:[sID = 1]]) ifTrue:[
			"MS uses Unicode all others single byte"
			multiBytes := pID = 3.
			string := entry stringAt: storageOffset + offset length: length multiByte: multiBytes.
			"Put the name at the right location.
			Note: We prefer Macintosh strings about everything else."
			nID < strings size ifTrue:[
				(pID = 1 or:[(strings at: nID+1) = ''])
					ifTrue:[strings at: nID+1 put: string].
			].
		].
	].
	fontDescription setStrings: strings.! !

!TTFontReader methodsFor: 'processing' stamp: 'ar 11/1/1998 22:18'!
processSimpleGlyph: glyph contours: nContours from: entry

	| endPts  nPts iLength flags |
	endPts := Array new: nContours.
	1 to: nContours do:[:i| endPts at: i put: entry nextUShort].
	glyph initializeContours: nContours with: endPts.
	nPts := endPts last + 1.
	iLength := entry nextUShort. "instruction length"
	entry skip: iLength.
	flags := self getGlyphFlagsFrom: entry size: nPts.
	self readGlyphXCoords: entry glyph: glyph nContours: nContours flags: flags endPoints: endPts.
	self readGlyphYCoords: entry glyph: glyph nContours: nContours flags: flags endPoints: endPts.
	glyph buildContours.! !

!TTFontReader methodsFor: 'public' stamp: 'StephaneDucasse 11/7/2011 22:45'!
readFrom: aStream

	| fontData headerEntry maxProfileEntry nameEntry indexLocEntry charMapEntry glyphEntry horzHeaderEntry horzMetricsEntry kerningEntry glyphOffset cmap numHMetrics indexToLocFormat |

	"Read the raw font byte data"
	aStream binary.
	fontData := aStream contents asByteArray.
	fontDescription := TTFontDescription new.

	"Search the tables required to build the font"
	(headerEntry := self getTableDirEntry: 'head' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a header table'].
	(maxProfileEntry := self getTableDirEntry: 'maxp' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a maximum profile table'].
	(nameEntry := self getTableDirEntry: 'name' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a name table'].
	(indexLocEntry := self getTableDirEntry: 'loca' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a relocation table'].
	(charMapEntry := self getTableDirEntry: 'cmap' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a character map table'].
	(glyphEntry := self getTableDirEntry: 'glyf' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a glyph table'].
	(horzHeaderEntry := self getTableDirEntry: 'hhea' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a horizontal header table'].
	(horzMetricsEntry := self getTableDirEntry: 'hmtx' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a horizontal metrics table'].
	(kerningEntry := self getTableDirEntry: 'kern' from: fontData) == nil ifTrue:[
		self crTrace: 'This font does not have a kerning table'].


	"Process the data"
	indexToLocFormat := self processFontHeaderTable: headerEntry.
	self processMaximumProfileTable: maxProfileEntry.
	self processNamingTable: nameEntry.
	glyphOffset := self processIndexToLocationTable: indexLocEntry format: indexToLocFormat.
	cmap := self processCharacterMappingTable: charMapEntry.
	(cmap == nil or:[cmap value == nil])
		ifTrue:[^self error:'This font has no suitable character mappings'].
	self processGlyphDataTable: glyphEntry offsets: glyphOffset.
	numHMetrics := self processHorizontalHeaderTable: horzHeaderEntry.
	self processHorizontalMetricsTable: horzMetricsEntry length: numHMetrics.
	kerningEntry isNil 
		ifTrue:[kernPairs := #()]
		ifFalse:[self processKerningTable: kerningEntry].
	charMap := self processCharMap: cmap.
	fontDescription setGlyphs: glyphs mapping: charMap.
	fontDescription setKernPairs: kernPairs.
	^fontDescription! !

!TTFontReader methodsFor: 'private' stamp: 'ar 11/2/1998 01:36'!
readGlyphXCoords:entry glyph: glyph nContours: nContours flags: flags endPoints: endPts
	"Read the x coordinates for the given glyph from the font file."
	| startPoint endPoint flagBits xValue contour ttPoint |
	startPoint := 1.
	1 to: nContours do:[:i|
		contour := glyph contours at: i.
		"Get the end point"
		endPoint := (endPts at: i) + 1.
		"Store number of points"
		startPoint to: endPoint do:[:j|
			ttPoint := contour points at: (j - startPoint + 1).
			flagBits := flags at: j.
			"If bit zero in the flag is set then this point is an on-curve
			point, if not, then it is an off-curve point."
			(flagBits bitAnd: 1) = 1 
				ifTrue:[ ttPoint type: #OnCurve]
				ifFalse:[ttPoint type: #OffCurve].
			"First we check to see if bit one is set.  This would indicate that
			the corresponding coordinate data in the table is 1 byte long.
			If the bit is not set, then the coordinate data is 2 bytes long."
			(flagBits bitAnd: 2) = 2 ifTrue:[ "one byte"
				xValue := entry nextByte.
				xValue := (flagBits bitAnd: 16)=16 ifTrue:[xValue] ifFalse:[xValue negated].
				ttPoint x: xValue.
			] ifFalse:[ "two byte"
				"If bit four is set, then this coordinate is the same as the
				last one, so the relative offset (of zero) is stored.  If bit
				is not set, then read in two bytes and store it as a signed value."
				(flagBits bitAnd: 16) = 16 ifTrue:[ ttPoint x: 0 ]
				ifFalse:[
					xValue := entry nextShort.
					ttPoint x: xValue]]].
		startPoint := endPoint + 1]! !

!TTFontReader methodsFor: 'private' stamp: 'ar 11/2/1998 01:37'!
readGlyphYCoords:entry glyph: glyph nContours: nContours flags: flags endPoints: endPts
	"Read the y coordinates for the given glyph from the font file."
	| startPoint endPoint flagBits yValue contour ttPoint |
	startPoint := 1.
	1 to: nContours do:[:i|
		contour := glyph contours at: i.
		"Get the end point"
		endPoint := (endPts at: i) + 1.
		"Store number of points"
		startPoint to: endPoint do:[:j|
			ttPoint := contour points at: (j - startPoint + 1).
			flagBits := flags at: j.
			"Check if this value one or two byte encoded"
			(flagBits bitAnd: 4) = 4 ifTrue:[ "one byte"
				yValue := entry nextByte.
				yValue := (flagBits bitAnd: 32)=32 ifTrue:[yValue] ifFalse:[yValue negated].
				ttPoint y: yValue.
			] ifFalse:[ "two byte"
				(flagBits bitAnd: 32) = 32 ifTrue:[ ttPoint y: 0 ]
				ifFalse:[
					yValue := entry nextShort.
					ttPoint y: yValue]]].
		startPoint := endPoint + 1]! !

!TTFontReader methodsFor: 'private' stamp: 'StephaneDucasse 11/7/2011 22:45'!
warn: aString
	self crTrace: aString.! !

!TTFontReader methodsFor: 'private' stamp: 'michael.rueger 2/5/2009 17:03'!
winToMac: index
	^ (index - 1) asCharacter unicodeToMacRoman charCode + 1! !

AbstractFont subclass: #TTCFont
	instanceVariableNames: 'ttcDescription pointSize foregroundColor cache derivatives fallbackFont height ascent descent colorToCacheMap'
	classVariableNames: 'GlyphCacheData GlyphCacheIndex GlyphCacheReady GlyphCacheSize Registry Scale ShutdownList'
	poolDictionaries: 'TextConstants'
	category: 'TrueType-Fonts'!

!TTCFont commentStamp: '<historical>' prior: 0!
I represent a font that uses TrueType derived glyph.  Upon a request for glyph for a character through a call to #formOf: (or #widthOf:), I first search corresponding glyph in the cache.  If there is not, it creates a 32bit depth form with the glyph.

  The cache is weakly held.  The entries are zapped at full GC.

Structure:
 ttcDescription	TTFontDescription -- The system data structure for a TrueType font data file.
 pointSize		Number -- Nominal Em size in points. Conversion to pixel sizes depends on the definition of TextStyle class>>pixelsPerInch.
 foregroundColor	Color -- So far, this font need to know the glyph color in cache.
 cache			WeakArray of <Color -> <Array(256) of glyph>>
 derivatives		Array -- stores the fonts in the same family but different emphasis.
!

TTCFont subclass: #LinedTTCFont
	instanceVariableNames: 'emphasis lineGlyph contourWidth'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TrueType-Fonts'!

!LinedTTCFont class methodsFor: 'as yet unclassified' stamp: 'ar 11/14/2006 15:18'!
fromTTCFont: aTTCFont emphasis: code

	| inst |
	inst := self new.
	inst ttcDescription: aTTCFont ttcDescription.
	inst pointSize: aTTCFont pointSize.
	inst emphasis: (aTTCFont emphasis bitOr: code).
	inst lineGlyph: (aTTCFont ttcDescription at: $_).

	^ inst.
! !

!LinedTTCFont methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:11'!
computeForm: char

	| ttGlyph scale |

	char = Character tab ifTrue: [^ super computeForm: char].

	"char = $U ifTrue: [self doOnlyOnce: [self halt]]."
	scale := self pixelSize asFloat / (ttcDescription ascender - ttcDescription descender).
	ttGlyph := ttcDescription at: char.
	^ ttGlyph asFormWithScale: scale ascender: ttcDescription ascender descender: ttcDescription descender fgColor: foregroundColor bgColor: Color transparent depth: self depth replaceColor: false lineGlyph: lineGlyph lingGlyphWidth: contourWidth emphasis: emphasis! !

!LinedTTCFont methodsFor: 'as yet unclassified' stamp: 'yo 5/6/2004 19:56'!
emphasis

	^ emphasis.
! !

!LinedTTCFont methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:11'!
emphasis: code

	emphasis := code.
! !

!LinedTTCFont methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:11'!
lineGlyph: aGlyph

	lineGlyph := aGlyph.
	contourWidth := aGlyph calculateWidth.
! !

TTCFont subclass: #MultiTTCFont
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TrueType-Fonts'!

!MultiTTCFont class methodsFor: 'as yet unclassified' stamp: 'yo 12/10/2002 18:34'!
cacheAllNil
"
	self cacheAllNil
"
	self allInstances do: [:inst |
		inst cache do: [:e |
			e third ifNotNil: [^ false].
		].
	].

	^ true.
! !

!MultiTTCFont methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:11'!
access: char at: index

	| wcache entry |
	wcache := self cache.
	entry := wcache at: index.
	wcache replaceFrom: index to: wcache size - 1 with: wcache startingAt: index + 1.
	wcache at: wcache size put: entry.
! !

!MultiTTCFont methodsFor: 'as yet unclassified' stamp: 'StephaneDucasse 5/6/2010 08:47'!
at: char put: form

	| wcache |
	wcache := self cache.
	wcache replaceFrom: 1 to: wcache size - 1 with: wcache startingAt: 2.
	wcache at: wcache size
		put: (Array with: char asciiValue with: foregroundColor with: form).
	^ form
! !

!MultiTTCFont methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:12'!
formOf: char

	| newForm |
	self hasCached: char ifTrue: [:form :index |
		self access: char at: index.
		^ form.
	].

	newForm := self computeForm: char.
	self at: char put: newForm.
	^ newForm.
! !

!MultiTTCFont methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:12'!
glyphInfoOf: char into: glyphInfoArray

	| newForm |
	self hasCached: char ifTrue: [:form :index |
		self access: char at: index.
		glyphInfoArray at: 1 put: form;
			at: 2 put: 0;
			at: 3 put: form width;
			at: 4 put: (self ascentOf: char);
			at: 5 put: self.
		^ glyphInfoArray.
	].

	newForm := self computeForm: char.
	self at: char put: newForm.

	glyphInfoArray at: 1 put: newForm;
		at: 2 put: 0;
		at: 3 put: newForm width;
		at: 4 put: (self ascentOf: char);
		at: 5 put: self.
	^ glyphInfoArray.
! !

!MultiTTCFont methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:12'!
hasCached: char ifTrue: twoArgBlock

	| value elem |
	value := char asciiValue.

	self cache size to: 1 by: -1 do: [:i |
		elem := self cache at: i.
		(elem first = value and: [elem second = foregroundColor]) ifTrue: [
			^ twoArgBlock value: elem third value: i.
		].
	].
	^ false.
! !

!MultiTTCFont methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:12'!
widthOf: char

	"This method cannot use #formOf: because formOf: discriminates the color and causes unnecessary bitmap creation."

	| newForm |
	self hasCached: char ifTrue: [:form :index |
		self access: char at: index.
		^ form width.
	].

	newForm := self computeForm: char.
	self at: char put: newForm.
	^ newForm width.

! !

!TTCFont class methodsFor: 'objects from disk' stamp: 'nk 6/25/2003 13:42'!
classVersion
	"Version 0 had pixelSize; version 1 changed it to pointSize"
	^1! !

!TTCFont class methodsFor: 'instance creation' stamp: 'yo 6/23/2003 18:44'!
family: f size: s

	^ self allInstances detect: [:a | a familyName = f and: [a pointSize = s]] ifNone: [nil].
! !

!TTCFont class methodsFor: 'instance creation' stamp: 'BenjaminVanRyseghem 11/28/2010 19:02'!
familyName: n pointSize: s emphasis: code

	"(TTCFont familyName: 'BitstreamVeraSans' pointSize: 12 emphasis: 0)"
	| t ret index |
	t := self registry at: n asSymbol ifAbsent: [#()].
	t isEmpty ifTrue: [
		t := (TextSharedInformation at: #DefaultTextStyle) fontArray.
		ret := t first.
		ret pointSize >= s ifTrue: [^ ret emphasis: code].
		index := 2.
		[index <= t size and: [(t at: index) pointSize <= s]] whileTrue: [
			ret := t at: index.
			index := index + 1.
		].
		^ ret emphasis: code.
	].
	^ ((TextStyle named: n) addNewFontSize: s) emphasis: code.
! !

!TTCFont class methodsFor: 'file list services' stamp: 'nk 7/16/2003 15:40'!
fileReaderServicesForFile: fullName suffix: suffix 
	^(suffix = 'ttf')  | (suffix = '*') 
		ifTrue: [ self services ]
		ifFalse: [ #() ]! !

!TTCFont class methodsFor: 'instance creation' stamp: 'sd 2/4/2008 21:20'!
getExistings: fontArray

	| result em |
	result := OrderedCollection new.
	result add: fontArray.
	1 to: 3 do: [:i |
		em := (fontArray collect: [:f | f emphasized: i]).
		(em at: 1) ~= (fontArray at: 1) ifTrue: [
			result add: em.
		].
	].
	^ result asArray.
! !

!TTCFont class methodsFor: 'initialization' stamp: 'IgorStasenko 4/26/2011 16:18'!
initialize
"
	self initialize
"

	| tt |
	self allSubInstancesDo:[:fnt| fnt flushCache].
	GlyphCacheSize := 512.
	GlyphCacheData := Array new: GlyphCacheSize.
	GlyphCacheIndex := 0.
	GlyphCacheReady := true.
	
	tt := TTFontDescription default.
	tt ifNotNil: [self newTextStyleFromTT: tt].

	(Smalltalk tools fileList respondsTo: #registerFileReader:) ifTrue: [
		Smalltalk tools fileList registerFileReader: self
	].

	Smalltalk addToShutDownList: self.! !

!TTCFont class methodsFor: 'other' stamp: 'yo 6/23/2003 19:46'!
isCacheAllNil
"
	self cacheAllNil
"
	self allInstances do: [:inst |
		inst cache do: [:e |
			e ifNotNil: [^ false].
		].
	].

	^ true.
! !

!TTCFont class methodsFor: 'instance creation' stamp: 'nice 1/5/2010 15:59'!
newTextStyleFromTT: description 
	"Create a new TextStyle from specified TTFontDescription instance."

	| array |
	array := self pointSizes collect: 
					[:pt | | f | 
					f := self new.
					f ttcDescription: description.
					f pointSize: pt].
	^self reorganizeForNewFontArray: array name: array first name asSymbol! !

!TTCFont class methodsFor: 'instance creation' stamp: 'sd 2/4/2008 21:20'!
newTextStyleFromTTFile: fileName
	"Create a new TextStyle from specified file name.  On certain versions of Windows, you can evaluate following to get Arial font into the image.  On other platforms, wait and see someone implements the support code for FontPlugin then we can start relying on the generic font lookup mechanism.
	TTCFontReader encodingTag: 0.
	self newTextStyleFromTTFile: 'C:\WINDOWS\Fonts\symbol.TTF'.
	"

	| description |
	description := TTFontDescription addFromTTFile: fileName.
	^ self newTextStyleFromTT: description.
! !

!TTCFont class methodsFor: 'instance creation' stamp: 'sd 2/4/2008 21:20'!
newTextStyleFromTTStream: readStream
"
"

	| description |
	description := TTFontDescription addFromTTStream: readStream.
	^ self newTextStyleFromTT: description.
! !

!TTCFont class methodsFor: 'other' stamp: 'yo 6/23/2003 20:18'!
pointSizes

	"The default sizes that are created when a TextStyle is created.  You can add new sizes by the new-size feature."
	^ #(9 12 15 24 36).
! !

!TTCFont class methodsFor: 'other' stamp: 'ar 11/14/2006 15:19'!
recreateCache
"
	self recreateCache.
"
	self allSubInstances do: [:inst | inst flushCache].
	Smalltalk garbageCollect.
! !

!TTCFont class methodsFor: 'other' stamp: 'yo 3/17/2005 10:52'!
register: anObject at: symbolName

	self registry at: symbolName put: anObject.
! !

!TTCFont class methodsFor: 'other' stamp: 'yo 3/17/2005 11:11'!
registerAll
"
	TTCFont registerAll
"

	TextStyle allInstancesDo: [:e |
		(e fontArray first isMemberOf: TTCFont) ifTrue: [
			self register: e fontArray at: e fontArray first familyName asSymbol.
		].
	].
! !

!TTCFont class methodsFor: 'other' stamp: 'NorbertHartl 6/13/2008 11:45'!
registry

	^ Registry isNil
		ifTrue: [Registry := IdentityDictionary new]
		ifFalse: [Registry].
! !

!TTCFont class methodsFor: 'other' stamp: 'yo 5/7/2004 08:09'!
removeAllDerivatives
"
	self removeAllDerivatives
"

	self allInstances do: [:s |
		s textStyle ifNotNil: [
			s textStyle fontArray do: [:f |
				f derivativeFont: nil at: 0.
			].
		].
	].
! !

!TTCFont class methodsFor: 'other' stamp: 'BenjaminVanRyseghem 11/28/2010 19:02'!
removeStyleName: aString

	TextSharedInformation removeKey: aString asSymbol ifAbsent: [].
	TTFontDescription removeDescriptionNamed: aString asString.
! !

!TTCFont class methodsFor: 'instance creation' stamp: 'BenjaminVanRyseghem 11/28/2010 19:02'!
reorganizeForNewFontArray: array name: styleName

	| style existings regular altName |
	(TextSharedInformation includesKey: styleName) ifFalse: [
		TextSharedInformation at: styleName put: (TextStyle fontArray: array).
		^ TextSharedInformation at: styleName.
	].
 
	"There is a text style with the name I want to use.  See if it is a TTC font..."
	style := TextSharedInformation at: styleName.
	style isTTCStyle ifFalse: [
		altName := ((array at: 1) name, 'TT') asSymbol.
		^ self reorganizeForNewFontArray: array name: altName.
	].

	existings := (self getExistings: style fontArray), (Array with: array).
	regular := existings detect: [:e | (e at: 1) isRegular] ifNone: [existings at: 1].

	regular do: [:r |
		r addLined: r.
	].

	"The existing array may be different in size than the new one."
	existings do: [:e |
		(e at: 1) isRegular ifFalse: [
			regular do: [ :r | | f |
				f := e detect: [ :ea | ea pointSize = r pointSize ] ifNone: [ ].
				f ifNotNil: [ r derivativeFont: f ].
			].
		].
	].

	style newFontArray: regular.
	self register: regular at: styleName.
	self recreateCache.	
	^ style.
! !

!TTCFont class methodsFor: 'other' stamp: 'nice 1/5/2010 15:59'!
repairBadSizes
	"There was a bug that would cause the TTCFonts to generate incorrectly sized glyphs.
	By looking at the dimensions of cached forms,
	we can tell whether the incorrect height logic was used.
	If it was, change the point size of the font and its derivatives.
	
	Note that this is probably pointless to call after the new code has been loaded; it's here for documentation (it should be called from the CS preamble instead)."

	"TTCFont repairBadSizes"
	| repaired |
	repaired := OrderedCollection new.
	TTCFont allInstancesDo: [ :font | | computedScale cached newPointSize description desiredScale |
		cached := (font cache copyFrom: $A asciiValue + 1 to: $z asciiValue + 1)
			detect: [ :f | f notNil ] ifNone: [].
		cached := cached ifNil: [  font formOf: $A ] ifNotNil: [ cached value ].
		description := font ttcDescription.
		desiredScale := cached height asFloat / (description ascender - description descender).
		computedScale := font pixelSize asFloat / font ttcDescription unitsPerEm.
		(((computedScale / desiredScale) - 1.0 * cached height) abs < 1.0) ifFalse: [
			newPointSize := (font pointSize * desiredScale / computedScale) rounded.
			font pointSize: newPointSize; flushCache.
			repaired add: font.
			font derivativeFonts do: [ :df | df ifNotNil: [
				df pointSize: newPointSize; flushCache.
				repaired add: df. ]].
		].
	].
	repaired isEmpty ifFalse: [ repaired asArray inspect ].
! !

!TTCFont class methodsFor: 'other' stamp: 'sd 2/4/2008 21:20'!
scale: anObject

	Scale := anObject.
! !

!TTCFont class methodsFor: 'file list services' stamp: 'tbn 8/11/2010 10:21'!
serviceInstallTrueTypeFontStyle
	"Return a service to install a true type font as a text style"

	^ SimpleServiceEntry
		provider: self
		label: 'Install ttf style'
		selector: #newTextStyleFromTTFile: 
		description: 'Install a true type font as a text style'
		buttonLabel: 'Install TTF'! !

!TTCFont class methodsFor: 'file list services' stamp: ''!
services
	"Return a set of services for use in FileList"

	^ Array with: self serviceInstallTrueTypeFontStyle! !

!TTCFont class methodsFor: 'initialization' stamp: 'ar 11/14/2006 15:28'!
shutDown
	"Flush the glyph cache"
	GlyphCacheData atAllPut: nil.
	GlyphCacheIndex := 0.
	ShutdownList ifNotNil:[ShutdownList do:[:fnt| fnt flushCache]].
	ShutdownList := WeakSet new.
! !

!TTCFont class methodsFor: 'initialization' stamp: 'IgorStasenko 4/26/2011 16:18'!
unload

	(Smalltalk tools fileList respondsTo: #unregisterFileReader:) ifTrue: [
		Smalltalk tools fileList unregisterFileReader: self
	]! !

!TTCFont class methodsFor: 'other' stamp: 'yo 3/17/2005 10:53'!
unregister: symbolName

	self registry removeKey: symbolName ifAbsent: [].
! !

!TTCFont class methodsFor: 'other' stamp: 'yo 6/23/2003 19:50'!
version

	^ '6.0'.
! !

!TTCFont methodsFor: 'private' stamp: 'yo 5/7/2004 12:30'!
addLined

	self addLined: self.
	self derivativeFonts do: [:e |
		e ifNotNil: [self addLined: e].
	].
! !

!TTCFont methodsFor: 'private' stamp: 'sd 2/4/2008 21:20'!
addLined: aTTCFont

	| l |
	l := LinedTTCFont fromTTCFont: aTTCFont emphasis: 4.
	self derivativeFont: l at: l emphasis.

	l := LinedTTCFont fromTTCFont: aTTCFont emphasis: 16.
	self derivativeFont: l at: l emphasis.

	l := LinedTTCFont fromTTCFont: aTTCFont emphasis: 20.
	self derivativeFont: l at: l emphasis.
! !

!TTCFont methodsFor: 'accessing' stamp: 'ar 11/14/2006 15:35'!
ascent
	ascent ifNil:[ascent := ttcDescription ascender * self pixelSize // (ttcDescription ascender - ttcDescription descender) * Scale y].
	^ (fallbackFont notNil
			and: [fallbackFont ascent > ascent])
		ifTrue: [fallbackFont ascent]
		ifFalse: [ascent]! !

!TTCFont methodsFor: 'private' stamp: 'StephaneDucasse 5/6/2010 08:50'!
at: char put: form
	| assoc |
	assoc := foregroundColor -> form.
	GlyphCacheData at: (GlyphCacheIndex := GlyphCacheIndex \\ GlyphCacheSize + 1) put: assoc.
	cache at: (char asInteger + 1) put: assoc.
	^ form! !

!TTCFont methodsFor: 'friend' stamp: 'ar 11/14/2006 15:19'!
cache
	^cache! !

!TTCFont methodsFor: 'character shapes' stamp: 'nk 11/3/2004 10:02'!
characterFormAt: character 
	"Answer a Form copied out of the glyphs for the argument,  
	character. Use a cached copy if possible."

	^self formOf: character! !

!TTCFont methodsFor: 'private' stamp: 'sd 2/4/2008 21:20'!
computeForm: char

	| ttGlyph scale |
	scale := self pixelSize asFloat / (ttcDescription ascender - ttcDescription descender).
	Scale ifNotNil: [scale := Scale * scale].
	ttGlyph := ttcDescription at: (char isCharacter ifTrue: [char charCode] ifFalse: [char]).
	^ ttGlyph asFormWithScale: scale ascender: ttcDescription ascender descender: ttcDescription descender fgColor: foregroundColor bgColor: Color transparent depth: self depth.
! !

!TTCFont methodsFor: 'copying' stamp: 'yo 12/10/2002 16:35'!
copy

	^ self.
! !

!TTCFont methodsFor: 'copying' stamp: 'yo 12/10/2002 16:35'!
deepCopy

	^ self.
! !

!TTCFont methodsFor: 'public' stamp: 'yo 12/10/2002 16:35'!
depth

	^ 32.
! !

!TTCFont methodsFor: 'friend' stamp: 'sd 2/4/2008 21:20'!
derivativeFont: aTTCFont

	| index |
	index := self indexOfSubfamilyName: (aTTCFont subfamilyName).
	index < 1 ifTrue: [
		^ self "inform: 'unknown sub family name.  This font will be skipped'".
	].

	self derivativeFont: aTTCFont at: index.

	self addLined: aTTCFont.
! !

!TTCFont methodsFor: 'friend' stamp: 'sd 2/4/2008 21:20'!
derivativeFont: aTTCFont at: index

	| newDeriv |
	aTTCFont ifNil: [derivatives := nil. ^ self].
	derivatives ifNil: [derivatives := Array new: 32].
	derivatives size < 32 ifTrue: [
		newDeriv := Array new: 32.
		newDeriv replaceFrom: 1 to: derivatives size with: derivatives.
		derivatives := newDeriv.
	].
	derivatives at: index put: aTTCFont.
! !

!TTCFont methodsFor: 'friend' stamp: 'yo 5/6/2004 19:54'!
derivativeFontArray

	^ derivatives.
! !

!TTCFont methodsFor: 'friend' stamp: 'nk 9/1/2004 13:01'!
derivativeFonts

	derivatives ifNil: [^ #()].
	^derivatives copyWithout: nil! !

!TTCFont methodsFor: 'accessing' stamp: 'ar 11/14/2006 15:43'!
descent
	"One is added to make sure the gap between lines is filled.  If we don't add, multi line selection in a text pane look ugly."
	^descent ifNil:[descent := (ttcDescription descender * self pixelSize // (ttcDescription descender - ttcDescription ascender)) * Scale y + 1].
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 11/30/2002 22:39'!
descentKern

	^ 0.
! !

!TTCFont methodsFor: 'friend' stamp: 'yo 1/7/2005 12:04'!
displayString: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta 

	^ self displayString: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta baselineY: aPoint y + self ascent.
! !

!TTCFont methodsFor: 'friend' stamp: 'ar 11/14/2006 15:40'!
displayString: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta baselineY: baselineY

	| form glyphInfo destX destY hereX nextX actualFont |
	destX := aPoint x.
	glyphInfo := Array new: 5.
	startIndex to: stopIndex do: [:charIndex |
		self glyphInfoOf: (aString at: charIndex) into: glyphInfo.
		form := glyphInfo at: 1.
		hereX := glyphInfo at: 2.
		nextX := glyphInfo at: 3.
		(actualFont := glyphInfo at: 5) ==  aBitBlt lastFont
			ifFalse: [actualFont installOn: aBitBlt].
		destY := baselineY - (glyphInfo at: 4). 
		aBitBlt sourceForm: form.
		aBitBlt destX: destX.
		aBitBlt destY: destY.
		aBitBlt sourceX: hereX; sourceY: 0.
		aBitBlt width: nextX - hereX.
		aBitBlt height: form height.
		aBitBlt copyBits.
		destX := destX + (nextX - hereX) + kernDelta.
	].
	^ destX @ destY
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 5/6/2004 19:25'!
emphasis
	"Answer the emphasis code (0 to 3) corresponding to my subfamily name"
	^self indexOfSubfamilyName: self subfamilyName

! !

!TTCFont methodsFor: 'accessing' stamp: 'nk 5/26/2003 20:13'!
emphasis: code

	code > 3 ifTrue: [^ self].
	code = 0 ifTrue: [^ self].
	derivatives isNil ifTrue: [^ self].
	^ (derivatives at: code) ifNil: [self].
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 8/1/2005 12:14'!
emphasized: code

	code = 0 ifTrue: [^ self].
	derivatives ifNil: [^ self].
	(((code bitAnd: 20) ~= 0) and: [
		derivatives size < code or: [(derivatives at: code) isNil]]) ifTrue: [
		self addLined.
	].
	^ (derivatives at: code) ifNil: [self].
! !

!TTCFont methodsFor: 'accessing' stamp: 'sd 2/4/2008 21:20'!
fallbackFont
	^ fallbackFont
		ifNil: [fallbackFont := FixedFaceFont new errorFont fontSize: self height]! !

!TTCFont methodsFor: 'accessing' stamp: 'sd 2/4/2008 21:20'!
fallbackFont: aFontSetOrNil

	fallbackFont := aFontSetOrNil.
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 12/10/2002 17:08'!
familyName

	^ ttcDescription name.
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 11/30/2002 22:39'!
familySizeFace

	^ Array
		with: self familyName
		with: self height
		with: 0.
! !

!TTCFont methodsFor: 'initialize' stamp: 'ar 11/14/2006 15:17'!
flushCache
	"Flush the cache of this font"
	cache := foregroundColor := colorToCacheMap := nil.! !

!TTCFont methodsFor: 'accessing' stamp: 'dgd 8/17/2004 22:10'!
fontNameWithPointSize
	^ self name withoutTrailingDigits , ' ' , self pointSize printString! !

!TTCFont methodsFor: 'public' stamp: 'yo 11/16/2002 00:59'!
foregroundColor

	^ foregroundColor.
! !

!TTCFont methodsFor: 'initialize' stamp: 'ar 11/14/2006 15:27'!
foregroundColor: fgColor
	"Install the given foreground color"
	foregroundColor = fgColor ifFalse:[
		foregroundColor := fgColor.
		colorToCacheMap ifNil:[colorToCacheMap := Dictionary new].
		cache := colorToCacheMap at: fgColor ifAbsentPut:[WeakArray new: 256].
		ShutdownList ifNotNil:[ShutdownList add: self].
	].
! !

!TTCFont methodsFor: 'private' stamp: 'ar 11/14/2006 16:03'!
formOf: char

	| code form |
	char charCode > 255
		ifTrue: [^ self fallbackFont formOf: char].

	cache ifNil:[self foregroundColor: Color black]. "make sure we have a cache"

	code := char charCode.
	form := cache at: (code + 1).
	form class == Association ifTrue:[^self computeForm: code]. "in midst of loading"
	form ifNil:[
		form := self computeForm: code.
		cache at: code+1 put: form.
		GlyphCacheData at: (GlyphCacheIndex := GlyphCacheIndex \\ GlyphCacheSize + 1) put: form.
	].
	^form
! !

!TTCFont methodsFor: 'private' stamp: 'ar 11/14/2006 15:43'!
glyphInfoOf: aCharacter into: glyphInfoArray
	"Answer the width of the argument as a character in the receiver."

	| form |
	(self hasGlyphOf: aCharacter) ifFalse: [
		^ self fallbackFont glyphInfoOf: aCharacter into: glyphInfoArray.
	].
	form := self formOf: aCharacter.
	glyphInfoArray at: 1 put: form;
		at: 2 put: 0;
		at: 3 put: form width;
		at: 4 put: ascent "(self ascentOf: aCharacter)";
		at: 5 put: self.
	^ glyphInfoArray.
! !

!TTCFont methodsFor: 'private' stamp: 'yo 1/6/2005 04:43'!
hasGlyphOf: aCharacter

	^ aCharacter charCode <= 255
! !

!TTCFont methodsFor: 'accessing' stamp: 'ar 11/14/2006 15:44'!
height
	"Answer my height in pixels. This will answer a Float."
	^height ifNil:[height := self pixelSize * Scale y]! !

!TTCFont methodsFor: 'private' stamp: 'nk 3/25/2004 17:01'!
indexOfSubfamilyName: aName
	| decoded |

	"decodeStyleName will consume all the modifiers and leave nothing if everything was recognized."
	decoded := TextStyle decodeStyleName: aName.
	decoded second isEmpty ifTrue: [ ^decoded first ].

	"If you get a halt here - please add the missing synonym to the lookup table in TextStyle>>decodeStyleName: ."
	
	self error: 'please add the missing synonym ', aName, ' to the lookup table in TextStyle>>decodeStyleName:'.

	^0.! !

!TTCFont methodsFor: 'friend' stamp: 'alain.plantec 5/28/2009 11:09'!
initialize

	super initialize.
	foregroundColor := Color black.
! !

!TTCFont methodsFor: 'friend' stamp: 'yo 1/6/2005 21:59'!
installOn: aDisplayContext

	^aDisplayContext installTTCFont: self.
! !

!TTCFont methodsFor: 'friend' stamp: 'ar 11/14/2006 15:26'!
installOn: aDisplayContext foregroundColor: fgColor backgroundColor: bgColor
	self foregroundColor: fgColor. "install color"
	aDisplayContext installTTCFont: self foregroundColor: foregroundColor backgroundColor: bgColor
! !

!TTCFont methodsFor: 'testing' stamp: 'yo 5/6/2004 19:25'!
isRegular
	"Answer true if I am a Regular/Roman font (i.e. not bold, etc.)"
	^ (self indexOfSubfamilyName: (self subfamilyName)) = 0.
! !

!TTCFont methodsFor: 'testing' stamp: 'nk 6/25/2003 12:55'!
isTTCFont
	^true! !

!TTCFont methodsFor: 'accessing' stamp: 'dgd 12/11/2003 12:47'!
lineGrid
	"Answer the relative space between lines"
	^ self ascent + self descent! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 11/16/2002 01:00'!
maxAscii

	^ ttcDescription size.
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 11/16/2002 01:00'!
minAscii

	^ 0.
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 11/16/2002 01:00'!
name

	^ ttcDescription name.
! !

!TTCFont methodsFor: 'accessing' stamp: 'nk 4/2/2004 11:27'!
pixelSize
	"Make sure that we don't return a Fraction"
	^ TextStyle pointsToPixels: pointSize! !

!TTCFont methodsFor: 'accessing' stamp: 'nk 4/2/2004 11:27'!
pixelSize: aNumber
	"Make sure that we don't return a Fraction"
	self pointSize: (TextStyle pixelsToPoints: aNumber) rounded.
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 6/23/2003 18:39'!
pointSize

	^ pointSize.
! !

!TTCFont methodsFor: 'accessing' stamp: 'nk 7/18/2004 15:32'!
pointSize: aNumber

	self privatePointSize: aNumber.
	derivatives ifNotNil: [ derivatives do: [ :f | f ifNotNil: [ f privatePointSize: aNumber ]]].
! !

!TTCFont methodsFor: 'printing' stamp: 'yo 5/6/2004 19:25'!
printOn: aStream
	aStream nextPutAll: 'TTCFont(';
		nextPutAll: self familyName; space;
		print: self pointSize; space;
		nextPutAll: self subfamilyName;
		nextPut: $)! !

!TTCFont methodsFor: 'accessing' stamp: 'nk 7/18/2004 15:31'!
privatePointSize: aNumber 
	pointSize = aNumber
		ifFalse: [pointSize := aNumber.
			self flushCache]! !

!TTCFont methodsFor: 'caching' stamp: 'nk 3/15/2004 18:52'!
releaseCachedState
	self flushCache.! !

!TTCFont methodsFor: 'caching' stamp: 'yo 6/30/2004 14:49'!
reset
! !

!TTCFont methodsFor: 'private' stamp: 'nk 4/1/2004 09:15'!
scale

	^ self pixelSize / ttcDescription unitsPerEm
! !

!TTCFont methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:20'!
setupDefaultFallbackFont

	| fonts f |
	fonts := TextStyle default fontArray.
	f := fonts first.
	1 to: fonts size do: [:i |
		self height > (fonts at: i) height ifTrue: [f := fonts at: i].
	].
	self fallbackFont: f.
	self reset.

! !

!TTCFont methodsFor: 'friend' stamp: 'sd 2/4/2008 21:20'!
setupDefaultFallbackFontTo: aTextStyleOrNil
"
	TTCFont allInstances do: [:i | i setupDefaultFallbackFontTo: (TextStyle named: 'MultiMSMincho')].
"

	| fonts f |
	aTextStyleOrNil ifNil: [
		self fallbackFont: nil.
		^ self.
	].
	fonts := aTextStyleOrNil fontArray.
	(aTextStyleOrNil defaultFont familyName endsWith: self familyName) ifTrue: [fallbackFont := nil. ^ self].

	f := fonts first.
	1 to: fonts size do: [:i |
		self height >= (fonts at: i) height ifTrue: [f := fonts at: i].
	].
	self fallbackFont: f.
	self reset.

! !

!TTCFont methodsFor: 'public' stamp: 'yo 12/10/2002 16:36'!
size

	^ ttcDescription size.
! !

!TTCFont methodsFor: 'private' stamp: 'yo 5/6/2004 19:23'!
subfamilyName

	^ ttcDescription subfamilyName.
! !

!TTCFont methodsFor: 'accessing' stamp: 'nk 6/17/2003 14:26'!
textStyle
	^ TextStyle actualTextStyles detect:
		[:aStyle | aStyle fontArray includes: self] ifNone: [nil]! !

!TTCFont methodsFor: 'friend' stamp: 'yo 11/16/2002 01:01'!
ttcDescription

	^ ttcDescription.
! !

!TTCFont methodsFor: 'friend' stamp: 'ar 11/14/2006 15:19'!
ttcDescription: aTTCDescription

	ttcDescription := aTTCDescription.
	self flushCache.
! !

!TTCFont methodsFor: 'copying' stamp: 'yo 12/10/2002 16:36'!
veryDeepCopyWith: deepCopier

	self flushCache.
	^ self.
! !

!TTCFont methodsFor: 'public' stamp: 'ar 11/14/2006 15:23'!
widthOf: aCharacter
	"This method cannot use #formOf: because formOf: discriminates the color and causes unnecessary bitmap creation."
	aCharacter charCode > 255 ifTrue: [
		fallbackFont ifNotNil: [^ fallbackFont widthOf: aCharacter].
		^ 1
	].
	^(self formOf: aCharacter) width! !

AbstractFont subclass: #TTCFontSet
	instanceVariableNames: 'fontArray foregroundColor'
	classVariableNames: 'Registry'
	poolDictionaries: 'TextConstants'
	category: 'TrueType-Fonts'!

!TTCFontSet class methodsFor: 'as yet unclassified' stamp: 'BenjaminVanRyseghem 11/28/2010 19:02'!
discardDefault
"
	self discardDefault
"
	| ttc |
	ttc := TTCFontDescription default.
	ttc ifNotNil: [
		TextSharedInformation removeKey: ttc name asSymbol ifAbsent: [].
	].! !

!TTCFontSet class methodsFor: 'as yet unclassified' stamp: 'BenjaminVanRyseghem 11/28/2010 19:02'!
familyName: n pointSize: s

	"(self familyName: 'MultiMSGothic' pointSize: 14) pointSize"
	| t ret index |
	t := self registry at: n asSymbol ifAbsent: [#()].
	t isEmpty ifTrue: [
		t := (TextSharedInformation at: #DefaultTextStyle) fontArray.
		ret := t first.
		ret pointSize >= s ifTrue: [^ ret].
		index := 2.
		[index <= t size and: [(t at: index) pointSize <= s]] whileTrue: [
			ret := t at: index.
			index := index + 1.
		].
		^ ret.
	].
	^ (TextStyle named: n) addNewFontSize: s.! !

!TTCFontSet class methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:20'!
initialize
"
	self initialize
"

	| tt |
	tt := TTCFontDescription default.
	tt ifNotNil: [self newTextStyleFromTT: tt].
! !

!TTCFontSet class methodsFor: 'as yet unclassified' stamp: 'yo 11/16/2002 01:18'!
newFontArray: anArray
 
	^super new initializeWithFontArray: anArray
! !

!TTCFontSet class methodsFor: 'as yet unclassified' stamp: 'BenjaminVanRyseghem 11/28/2010 19:02'!
newTextStyleFromTT: descriptionArray

	| array textStyle styleName arrayOfArray |

	arrayOfArray := self pointSizes collect: [:pt |
		descriptionArray collect: [:ttc | | f |
			ttc ifNil: [nil] ifNotNil: [
				f := (ttc size > 256)
					ifTrue: [MultiTTCFont new initialize]
					ifFalse: [TTCFont new initialize].
				f ttcDescription: ttc.
				f pointSize: pt.
			].
		].
	].

	array := arrayOfArray collect: [:fonts |
		self newFontArray: fonts.
	].

	styleName := (array at: 1) familyName asSymbol.
	textStyle := TextStyle fontArray: array.
	TextSharedInformation at: styleName put: textStyle.

	self register: array at: styleName.

	^ TextSharedInformation at: styleName.
! !

!TTCFontSet class methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:20'!
newTextStyleFromTTFile: fileName
"
	TTCFontReader encodingTag: JapaneseEnvironment leadingChar.
	self newTextStyleFromTTFile: 'C:\WINDOWS\Fonts\msmincho.TTC'

	TTCFontReader encodingTag: 0.
	self newTextStyleFromTTFile: 'C:\WINDOWS\Fonts\symbol.ttf'
"

	| description |
	description := TTCFontDescription addFromTTFile: fileName.
	^ self newTextStyleFromTT: description.
! !

!TTCFontSet class methodsFor: 'as yet unclassified' stamp: 'yo 11/30/2002 22:21'!
pointSizes

	^ TTCFont pointSizes.
! !

!TTCFontSet class methodsFor: 'as yet unclassified' stamp: 'yo 12/2/2004 12:50'!
register: anObject at: symbolName

	self registry at: symbolName put: anObject.
! !

!TTCFontSet class methodsFor: 'as yet unclassified' stamp: 'yo 12/2/2004 13:07'!
registry

	^ Registry isNil
		ifTrue: [Registry := IdentityDictionary new]
		ifFalse: [Registry].
! !

!TTCFontSet class methodsFor: 'as yet unclassified' stamp: 'BenjaminVanRyseghem 11/28/2010 19:02'!
removeStyleName: aString

	| style symName |
	symName := aString asSymbol.
	style := TextSharedInformation removeKey: symName ifAbsent: [].
	style ifNotNil: [self unregister: symName].
	TTCFontDescription removeDescriptionNamed: aString asString.
! !

!TTCFontSet class methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:20'!
setDefault
"
	self setDefault
"
	| tt |
	tt := TTCFontDescription default.
	tt ifNil: [TTCFontDescription setDefault].
	tt := TTCFontDescription default.
	tt ifNotNil: [self newTextStyleFromTT: tt].
! !

!TTCFontSet class methodsFor: 'as yet unclassified' stamp: 'yo 12/2/2004 12:49'!
unregister: symbolName

	self registry removeKey: symbolName ifAbsent: [].
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'yo 11/16/2002 01:15'!
ascent

	^ (fontArray at: 1) ascent.
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'JMM 11/4/2008 22:57'!
ascentOf: aCharacter

	^ (fontArray at: 1) ascentOf: aCharacter.
! !

!TTCFontSet methodsFor: 'as yet unclassified' stamp: 'yo 12/10/2002 18:28'!
depth

	^ (fontArray at: 1) depth.
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'yo 11/16/2002 01:16'!
descent

	^ (fontArray at: 1) descent.
! !

!TTCFontSet methodsFor: 'as yet unclassified' stamp: 'yo 11/16/2002 01:16'!
descentKern

	^ 0.
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'JMM 11/4/2008 23:29'!
descentOf: aChar

	^ (fontArray  at: 1) descentOf: aChar
! !

!TTCFontSet methodsFor: 'displaying' stamp: 'yo 1/7/2005 12:05'!
displayString: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta

	^ self displayString: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta baselineY: aPoint y + self ascent.
! !

!TTCFontSet methodsFor: 'displaying' stamp: 'nice 3/16/2010 20:13'!
displayString: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta baselineY: baselineY

	| destPoint font form encoding glyphInfo char charCode destY |
	destPoint := aPoint.
	glyphInfo := Array new: 5.
	startIndex to: stopIndex do: [:charIndex |
		char := aString at: charIndex.
		encoding := char leadingChar + 1.
		charCode := char charCode.
		font := fontArray at: encoding.
		((charCode between: font minAscii and: font maxAscii) not) ifTrue: [
			charCode := font maxAscii].
		self glyphInfoOf: char into: glyphInfo.
		form := glyphInfo at: 1.
		(glyphInfo fifth ~= aBitBlt lastFont) ifTrue: [
			glyphInfo fifth installOn: aBitBlt.
		].
		destY := baselineY - glyphInfo fourth. 
		aBitBlt
			sourceForm: form;
			destX: destPoint x;
			destY: destY;
			sourceOrigin: 0 @ 0;
			width: form width;
			height: form height;
			copyBits.
		destPoint := destPoint x + (form width + kernDelta) @ destPoint y.
	].
	^ destPoint.
! !

!TTCFontSet methodsFor: 'as yet unclassified' stamp: 'JMM 11/4/2008 23:29'!
emphasis
	^ (fontArray  at: 1) emphasis! !

!TTCFontSet methodsFor: 'as yet unclassified' stamp: 'yo 11/16/2002 01:16'!
emphasized: code

! !

!TTCFontSet methodsFor: 'accessing' stamp: 'yo 12/10/2002 18:20'!
familyName

	^ 'Multi', (fontArray at: 1) familyName.
! !

!TTCFontSet methodsFor: 'as yet unclassified' stamp: 'JMM 11/4/2008 23:29'!
familySizeFace

	^ Array
		with: (fontArray  at: 1) name
		with: self height
		with: 0.
! !

!TTCFontSet methodsFor: 'as yet unclassified' stamp: 'yo 11/16/2002 01:16'!
fontArray

	^ fontArray
! !

!TTCFontSet methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:20'!
glyphInfoOf: aCharacter into: glyphInfoArray

	| index f code |
	index := aCharacter leadingChar + 1.
	fontArray size < index ifTrue: [^ self questionGlyphInfoInto: glyphInfoArray].
	(f := fontArray at: index) ifNil: [^ self questionGlyphInfoInto: glyphInfoArray].

	code := aCharacter charCode.
	((code between: f minAscii and: f maxAscii) not) ifTrue: [
		^ self questionGlyphInfoInto: glyphInfoArray.
	].
	f glyphInfoOf: aCharacter into: glyphInfoArray.
	glyphInfoArray at: 5 put: self.
	^ glyphInfoArray.
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'JMM 11/4/2008 23:29'!
height

	^(fontArray at: 1) pixelSize.
! !

!TTCFontSet methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:20'!
initializeWithFontArray: anArray

	fontArray := anArray.
	"name := anArray first name."
! !

!TTCFontSet methodsFor: 'as yet unclassified' stamp: 'yo 1/6/2005 22:00'!
installOn: aDisplayContext

	^aDisplayContext installTTCFont: self.
! !

!TTCFontSet methodsFor: 'displaying' stamp: 'sd 2/4/2008 21:20'!
installOn: aDisplayContext foregroundColor: fgColor backgroundColor: bgColor

	foregroundColor := fgColor.
	fontArray do: [:s | s ifNotNil: [s installOn: aDisplayContext foregroundColor: fgColor backgroundColor: bgColor]].
! !

!TTCFontSet methodsFor: 'testing' stamp: 'yo 2/12/2007 19:34'!
isFontSet

	^ true.
! !

!TTCFontSet methodsFor: 'testing' stamp: 'yo 12/29/2003 15:02'!
isTTCFont
	^true! !

!TTCFontSet methodsFor: 'accessing' stamp: 'JMM 11/4/2008 23:30'!
lineGrid

	^ (fontArray  at: 1) lineGrid.
! !

!TTCFontSet methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:20'!
maxAsciiFor: encoding

	| f |
	f := (fontArray at: encoding+1).
	f ifNotNil: [^ f maxAscii].
	^ 0.
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'JMM 11/4/2008 23:30'!
pointSize

	^ (fontArray  at: 1) pixelSize * 72 // 96.
! !

!TTCFontSet methodsFor: 'as yet unclassified' stamp: 'yo 11/16/2002 01:27'!
pointSizes

	^ self class pointSizes.
! !

!TTCFontSet methodsFor: 'as yet unclassified' stamp: 'sd 2/4/2008 21:20'!
questionGlyphInfoInto: glyphInfoArray

	| f form |
	f := fontArray at: 1.
	form := f formOf: $?.
	glyphInfoArray at: 1 put: form;
		at: 2 put: 0;
		at: 3 put: form width;
		at: 4 put: self.
	^ glyphInfoArray.
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'yo 8/5/2003 15:31'!
textStyle

	^ TextStyle actualTextStyles
		detect: [:aStyle | (aStyle fontArray collect: [:s | s name]) includes: self name]
		ifNone: [].
! !

!TTCFontSet methodsFor: 'as yet unclassified' stamp: 'JMM 11/4/2008 23:30'!
ttcDescription
	^ (fontArray  at: 1) ttcDescription! !

!TTCFontSet methodsFor: 'measuring' stamp: 'sd 2/4/2008 21:20'!
widthOf: aCharacter

	| encoding |
	encoding := aCharacter leadingChar.
	^ (fontArray at: encoding + 1) widthOf: aCharacter.
! !
TTFontReader initialize!
TTFontDescription initialize!
TTCFontDescription initialize!
TTCFont initialize!
TTCFontSet initialize!
